#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""A collection of calibration data from SBS instruments.

These are calibration coefficients from specific SBS intstruments that are primarily used for test purposes.
"""

# Native imports
from dataclasses import dataclass

# Third-party imports

# Sea-Bird imports

# Internal imports


class TemperatureCoefficients:
    """
    Args:
        a0 (float): a0 calibration coefficient for the temperature sensor
        a1 (float): a1 calibration coefficient for the temperature sensor
        a2 (float): a2 calibration coefficient for the temperature sensor
        a3 (float): a3 calibration coefficient for the temperature sensor
    """

    def __init__(self, a0, a1, a2, a3):
        self.a0 = a0
        self.a1 = a1
        self.a2 = a2
        self.a3 = a3


class PressureCoefficients:
    """
    Args:
        pa0 (float): pa0 calibration coefficient for the pressure sensor
        pa1 (float): pa1 calibration coefficient for the pressure sensor
        pa2 (float): pa2 calibration coefficient for the pressure sensor
        ptca0 (float): ptca0 calibration coefficient for the pressure sensor
        ptca1 (float): ptca1 calibration coefficient for the pressure sensor
        ptca2 (float): ptca2 calibration coefficient for the pressure sensor
        ptcb0 (float): ptcb0 calibration coefficient for the pressure sensor
        ptcb1 (float): ptcb1 calibration coefficient for the pressure sensor
        ptcb2 (float): ptcb2 calibration coefficient for the pressure sensor
        ptempa0 (float): ptempa0 calibration coefficient for the pressure sensor
        ptempa1 (float): ptempa1 calibration coefficient for the pressure sensor
        ptempa2 (float): ptempa2 calibration coefficient for the pressure sensor
    """

    def __init__(
        self, pa0, pa1, pa2, ptca0, ptca1, ptca2, ptcb0, ptcb1, ptcb2, ptempa0, ptempa1, ptempa2
    ):
        self.pa0 = pa0
        self.pa1 = pa1
        self.pa2 = pa2
        self.ptca0 = ptca0
        self.ptca1 = ptca1
        self.ptca2 = ptca2
        self.ptcb0 = ptcb0
        self.ptcb1 = ptcb1
        self.ptcb2 = ptcb2
        self.ptempa0 = ptempa0
        self.ptempa1 = ptempa1
        self.ptempa2 = ptempa2


class ConductivityCoefficients:
    """
    Args:
        g (float): g calibration coefficient for the conductivity sensor
        h (float): h calibration coefficient for the conductivity sensor
        i (float): i calibration coefficient for the conductivity sensor
        j (float): j calibration coefficient for the conductivity sensor
        cpcor (float): cpcor calibration coefficient for the conductivity sensor
        ctcor (float): ctcor calibration coefficient for the conductivity sensor
        wbotc (float): Wien bridge oscillator temperature coefficient
            see the 37 Manual: https://www.seabird.com/asset-get.download.jsa?id=54627862348
    """

    def __init__(self, g, h, i, j, cpcor, ctcor, wbotc):
        self.g = g
        self.h = h
        self.i = i
        self.j = j
        self.cpcor = cpcor
        self.ctcor = ctcor
        self.wbotc = wbotc


class ECOCoefficients:
    """
    Args:
        slope (float): units/count for digital, units/V for analog
        offset (float): dark counts for digital, dark voltage for analog
    """

    def __init__(self, slope, offset):
        self.slope = slope
        self.offset = offset


class Oxygen63Coefficients:
    def __init__(self, a0, a1, a2, b0, b1, c0, c1, c2, e):
        self.a0 = a0
        self.a1 = a1
        self.a2 = a2
        self.b0 = b0
        self.b1 = b1
        self.c0 = c0
        self.c1 = c1
        self.c2 = c2
        self.e = e


class Thermistor63Coefficients:
    """
    Args:
        ta0 (float): calibration coefficient for the thermistor in the SBE63 sensor
        ta1 (float): calibration coefficient for the thermistor in the SBE63 sensor
        ta2 (float): calibration coefficient for the thermistor in the SBE63 sensor
        ta3 (float): calibration coefficient for the thermistor in the SBE63 sensor
    """

    def __init__(self, ta0, ta1, ta2, ta3):
        self.ta0 = ta0
        self.ta1 = ta1
        self.ta2 = ta2
        self.ta3 = ta3


class Oxygen43Coefficients:
    def __init__(self, soc, v_offset, tau_20, a, b, c, e, d0, d1, d2, h1, h2, h3):
        self.soc = soc
        self.v_offset = v_offset
        self.tau_20 = tau_20
        self.a = a
        self.b = b
        self.c = c
        self.e = e
        self.d0 = d0  # TODO: the sensor lists a d0 coefficient, but it doesn't seem to be used?
        self.d1 = d1
        self.d2 = d2
        self.h1 = h1
        self.h2 = h2
        self.h3 = h3


class PH18Coefficients:
    """
    Args:
        offset (float): calibration offset
        slope (float): calibration slope
    """

    def __init__(self, slope, offset):
        self.slope = slope
        self.offset = offset


class PARCoefficients:
    """
    Args:
        im (float): immersion coefficient
        a0 (float): calibration slope
        a1 (float): calibration offset
        multiplier (float): 1.0 for units of μEinsteins/m2 sec
    """

    def __init__(self, im, a0, a1, multiplier):
        self.im = im
        self.a0 = a0
        self.a1 = a1
        self.multiplier = multiplier


class PHSeaFETInternalCoefficients:
    """
    :param int_k0: Internal K0 calibration coefficient
    :param int_k2: Internal K2 calibration coefficient
    """

    def __init__(self, int_k0, int_k2):
        self.int_k0 = int_k0
        self.int_k2 = int_k2


class PHSeaFETExternalCoefficients:
    """
    :param ext_k0: External K0 calibration coefficient
    :param ext_k2: External K2 calibration coefficient
    :param f1: f(P) calibration coefficient
    :param f2: f(P) calibration coefficient
    :param f3: f(P) calibration coefficient
    :param f4: f(P) calibration coefficient
    :param f5: f(P) calibration coefficient
    :param f6: f(P) calibration coefficient
    """

    def __init__(self, ext_k0, ext_k2, f1, f2, f3, f4, f5, f6):
        self.ext_k0 = ext_k0
        self.ext_k2 = ext_k2
        self.f1 = f1
        self.f2 = f2
        self.f3 = f3
        self.f4 = f4
        self.f5 = f5
        self.f6 = f6



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""SBE variable name conversion to matlab friendly strings.

DESCRIPTION:
Converts the strings that define SBE variables to strings that exclude special characters. 

Gives units for each variable and defines whether it is a number or a text string.

All variables imported from SBEDataProcessing_7.26.4.pdf. There may be
additional variables added in future versions of Sea-Bird Data processing
modules that are not included here.

Functions:

    interpret_sbs_variable(sbs_var)

"""

# Native imports
import re

# Third-party imports

# Sea-Bird imports

# Internal imports


def interpret_sbs_variable(sbs_var):
    """Converts the strings that define SBE variables to strings that exclude special characters.

    Gives units for each variable and defines whether it is a number or a text string.

    Args:
        sbs_var (string): Name of the dataset

    Returns:
        list[str]: The name with special characters removed, the data format, and the units.
            kvar_name       =   converted, matlab code readable variable name
            kvar_format     =   defines whether variable is a float or a string
            kvar_units      =   variable units. Empty string if undefined.
    """

    # TEMPERATURE
    # primary sensor
    kvar_format = r"%f"

    if sbs_var in {"t090Cm", "t4990C", "tnc90C", "tv290C", "t090C"}:
        kvar_name = "t090C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"t090F", "t4990F", "tnc90F", "tv290F"}:
        kvar_name = "t090F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"t068C", "t4968C", "tnc68C", "tv268C"}:
        kvar_name = "t068C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"t068F", "t4968F", "tnc68F", "tv268F"}:
        kvar_name = "t068F"
        kvar_units = "ITS-68, deg F"
    elif sbs_var in {"t090"}:
        kvar_name = "t090"
        kvar_units = "SBE49, deg C"

    # secondary temperature
    elif sbs_var in {"t190C", "tnc290C"}:
        kvar_name = "t190C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"t190F", "tnc290F"}:
        kvar_name = "t190F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"t168C", "tnc268C"}:
        kvar_name = "t168C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"t168F", "tnc268F"}:
        kvar_name = "t168F"
        kvar_units = "ITS-68, deg F"

    # SBE38 REFERENCE TEMPERATURE
    # primary sensor
    elif sbs_var in {"t3890C", "t38_90C"}:
        kvar_name = "t03890C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"t3890F", "t38_90F"}:
        kvar_name = "t03890F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"t3868C", "t38_68C"}:
        kvar_name = "t03868C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"t3868F", "t38_68F"}:
        kvar_name = "t03868F"
        kvar_units = "ITS-68, deg F"
    # secondary sensor
    elif sbs_var in {"t3890C1"}:
        kvar_name = "t13890C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"t3890F1"}:
        kvar_name = "t13890F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"t3868C1"}:
        kvar_name = "t13868C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"t3868F1"}:
        kvar_name = "t13868F"
        kvar_units = "ITS-68, deg F"

    # TEMPERATURE DIFFERENCE
    elif sbs_var in {"T2-T190C"}:
        kvar_name = "tdiff90C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"T2-T190F"}:
        kvar_name = "tdiff90F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"T2-T168C"}:
        kvar_name = "tdiff68C"
        kvar_units = "IPTS-68, deg C"
    elif sbs_var in {"T2-T168F"}:
        kvar_name = "tdiff68F"
        kvar_units = "IPTS-68, deg F"

    # CONDUCTIVITY
    # primary conductivity
    elif sbs_var in {"c_S/m", "cond0S/m", "c0S/m"}:
        kvar_name = "c0Sm"
        kvar_units = "S/m"
    elif sbs_var in {"c_mS/cm", "cond0mS/cm", "c0mS/cm"}:
        kvar_name = "c0mScm"
        kvar_units = "mS/cm"
    elif sbs_var in {"c_uS/cm", "cond0uS/cm", "c0uS/cm"}:
        kvar_name = "c0uScm"
        kvar_units = "microS/cm"
    # secondary conductivity
    elif sbs_var in {"c1S/m"}:
        kvar_name = "c1Sm"
        kvar_units = "S/m"
    elif sbs_var in {"c1mS/cm"}:
        kvar_name = "c1mScm"
        kvar_units = "mS/cm"
    elif sbs_var in {"c1uS/cm"}:
        kvar_name = "c1uScm"
        kvar_units = "microS/cm"

    # CONDUCTIVITY DIFFERENCE
    elif sbs_var in {"C2-C1S/m"}:
        kvar_name = "cdiffSm"
        kvar_units = "S/m"
    elif sbs_var in {"C2-C1mS/cm"}:
        kvar_name = "cdiffmScm"
        kvar_units = "mS/cm"
    elif sbs_var in {"C2-C1uS/cm"}:
        kvar_name = "cdiffuScm"
        kvar_units = "microS/cm"

    # PRESSURE
    # user entered pressure values
    elif sbs_var in {"pr"}:  # SBE 49
        kvar_name = "p"
        kvar_units = "db"
    elif sbs_var in {"prM"}:
        kvar_name = "pM"
        kvar_units = "db (user entered)"
    elif sbs_var in {"prE"}:
        kvar_name = "pE"
        kvar_units = "psi (user entered)"
    # digiquartz pressure
    elif sbs_var in {"prDM"}:
        kvar_name = "pm"
        kvar_units = "db"
    elif sbs_var in {"prDE"}:
        kvar_name = "pe"
        kvar_units = "psi"
    # strain gauge pressure
    elif sbs_var in {"prSM", "prdM"}:
        kvar_name = "pm"
        kvar_units = "db"
    elif sbs_var in {"prSE", "prdE"}:
        kvar_name = "pe"
        kvar_units = "psi"
    # pressure temperature
    elif sbs_var in {"ptempC"}:
        kvar_name = "ptempC"
        kvar_units = "deg C"
    elif sbs_var in {"ptempF"}:
        kvar_name = "ptempF"
        kvar_units = "deg F"
    # FGP Pressure
    elif sbs_var in {"fgp0"}:
        kvar_name = "fgp0"
        kvar_units = "KPa"
    elif sbs_var in {"fgp1"}:
        kvar_name = "fgp1"
        kvar_units = "KPa"
    elif sbs_var in {"fgp2"}:
        kvar_name = "fgp2"
        kvar_units = "KPa"
    elif sbs_var in {"fgp3"}:
        kvar_name = "fgp3"
        kvar_units = "KPa"
    elif sbs_var in {"fgp4"}:
        kvar_name = "fgp4"
        kvar_units = "KPa"
    elif sbs_var in {"fgp5"}:
        kvar_name = "fgp5"
        kvar_units = "KPa"
    elif sbs_var in {"fgp6"}:
        kvar_name = "fgp6"
        kvar_units = "KPa"
    elif sbs_var in {"fgp7"}:
        kvar_name = "fgp7"
        kvar_units = "KPa"
    # SBE 50
    elif sbs_var in {"pr50M"}:
        kvar_name = "p050m"
        kvar_units = "db"
    elif sbs_var in {"pr50E"}:
        kvar_name = "p050e"
        kvar_units = "psi"
    elif sbs_var in {"pr50M1"}:
        kvar_name = "p150m"
        kvar_units = "db"
    elif sbs_var in {"pr50E1"}:
        kvar_name = "p150e"
        kvar_units = "psi"

    # FREQUENCY OUTPUT
    elif sbs_var in {
        "f0",
        "f1",
        "f2",
        "f3",
        "f4",
        "f5",
        "f6",
        "f7",
        "f8",
        "f9",
        "f10",
        "f11",
        "f12",
        "f13",
        "f14",
        "f15",
        "f16",
        "f17",
        "f18",
        "f19",
        "f20",
        "f21",
        "f22",
        "f23",
        "f24",
        "f25",
        "f26",
        "f27",
        "f28",
        "f29",
        "f30",
        "f31",
        "f32",
        "f33",
        "f34",
        "f35",
        "f36",
    }:
        kvar_name = sbs_var
        kvar_units = "Hz"

    # VOLTAGE OUTPUT
    elif sbs_var in {
        "v0",
        "v1",
        "v2",
        "v3",
        "v4",
        "v5",
        "v6",
        "v7",
        "v8",
        "v9",
        "v10",
        "v11",
        "v12",
        "v13",
        "v14",
        "v15",
    }:
        kvar_name = sbs_var
        kvar_units = "V"

    # GTD-DO SENSORS
    # PRESSURE
    elif sbs_var in {"GTDDOP0", "GTDDOP1", "GTDDOPdiff"}:
        kvar_name = sbs_var
        kvar_units = "mb"
    # TEMPERATURE
    elif sbs_var in {"GTDDOT0", "GTDDOT1", "GTDDOTdiff"}:
        kvar_name = sbs_var
        kvar_units = "deg C"
    # GTD-N2 SENSOR
    # PRESSURE
    elif sbs_var in {"GTDN2P0", "GTDN2P1", "GTDN2Pdiff"}:
        kvar_name = sbs_var
        kvar_units = "mb"
    # TEMPERATURE
    elif sbs_var in {"GTDN2T0", "GTDN2T1", "GTDN2Tdiff"}:
        kvar_name = sbs_var
        kvar_units = "deg C"

    # TIME
    # elapsed time
    elif sbs_var in {"timeS"}:
        kvar_name = "timeS"
        kvar_units = "elapsed time seconds"
    elif sbs_var in {"timeM"}:
        kvar_name = "timeM"
        kvar_units = "elapsed time minutes"
    elif sbs_var in {"timeH"}:
        kvar_name = "timeH"
        kvar_units = "elapsed time hours"
    elif sbs_var in {"timeJ"}:
        kvar_name = "timeJ"
        kvar_units = "elapsed time Julian Days"
    elif sbs_var in {"timeN"}:
        kvar_name = "timeN"
        kvar_units = "From NMEA: seconds since January 1, 2000"
    elif sbs_var in {"timeQ"}:
        kvar_name = "timeQ"
        kvar_units = "From NMEA: seconds since January 1, 1970"
    elif sbs_var in {"timeK"}:
        kvar_name = "timeK"
        kvar_units = "SBE timestamp: seconds since January 1, 2000"
    elif sbs_var in {"timeJV2"}:  # julian days
        kvar_name = "jdays"
        kvar_units = "SBE timestamp: julian days"
    elif sbs_var in {"timeSCP"}:  # julian days
        kvar_name = "jdays"
        kvar_units = "SBE timestamp: julian days"
    elif sbs_var in {"timeY"}:
        kvar_name = "timeY"
        kvar_units = "Computer Time: seconds since January 1, 1970"

    # POSITION
    elif sbs_var in {"latitude"}:
        kvar_name = "lat"
        kvar_units = "from NMEA degrees"
    elif sbs_var in {"longitude"}:
        kvar_name = "lon"
        kvar_units = "from NMEA degrees"

    # DEPTH
    elif sbs_var in {"depSM"}:
        kvar_name = "depSM"
        kvar_units = "depth salt water, m"
    elif sbs_var in {"depSF"}:
        kvar_name = "depSF"
        kvar_units = "depth salt water, ft"
    elif sbs_var in {"depFM"}:
        kvar_name = "depFM"
        kvar_units = "depth fresh water, m"
    elif sbs_var in {"depFF"}:
        kvar_name = "depSM"
        kvar_units = "depth fresh water, ft"
    elif sbs_var in {"dNMEA"}:
        kvar_name = "depSM"
        kvar_units = "depth from NMEA salt water, m"

    # POTENTIAL TEMPERATURE
    # I use the abbreviation of the greek symbol theta to designate
    # potential temperature
    elif sbs_var in {"potemp090C"}:
        kvar_name = "th090C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"potemp090F"}:
        kvar_name = "th090F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"potemp068C"}:
        kvar_name = "th068C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"potemp068F"}:
        kvar_name = "th068F"
        kvar_units = "ITS-68, deg F"
    elif sbs_var in {"potemp190C"}:
        kvar_name = "th190C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"potemp190F"}:
        kvar_name = "th190F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"potemp168C"}:
        kvar_name = "th168C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"potemp168F"}:
        kvar_name = "th168F"
        kvar_units = "ITS-68, deg F"

    # POTENTIAL TEMPEATURE DIFFERENCE
    elif sbs_var in {"potemp90Cdiff"}:
        kvar_name = "th090Cdiff"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"potemp90Fdiff"}:
        kvar_name = "th090Fdiff"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"potemp68Cdiff"}:
        kvar_name = "th068Cdiff"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"potemp68Fdiff"}:
        kvar_name = "th068Fdiff"
        kvar_units = "ITS-68, deg F"

    # POTENTIAL TEMPEATURE ANOMALY
    elif sbs_var in {"tha090C"}:
        kvar_name = "tha090C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"tha090F"}:
        kvar_name = "tha090F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"tha068C"}:
        kvar_name = "tha068C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"tha068F"}:
        kvar_name = "tha068F"
        kvar_units = "ITS-68, deg F"
    elif sbs_var in {"tha190C"}:
        kvar_name = "tha190C"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"tha190F"}:
        kvar_name = "tha190F"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"tha168C"}:
        kvar_name = "tha168C"
        kvar_units = "ITS-68, deg C"
    elif sbs_var in {"tha168F"}:
        kvar_name = "tha168F"
        kvar_units = "ITS-68, deg F"
    elif sbs_var in {"pta090C"}:
        kvar_name = "pta090C"
        kvar_units = "ITS-90, deg C"

    # SALINITY
    elif sbs_var in {"sal00"}:
        kvar_name = "psal0"
        kvar_units = "PSU"
    elif sbs_var in {"sal11"}:
        kvar_name = "psal1"
        kvar_units = "PSU"
    elif sbs_var in {"secS-priS"}:
        kvar_name = "psaldiff"
        kvar_units = "PSU"

    # DENSITY
    # sg: potential density, based on greek letter sg
    # sgth: potential denisty sg calculated with potential temperature
    # theta (th)
    elif sbs_var in {"density00"}:
        kvar_name = "rho0"
        kvar_units = "density, kg/m^3"
    elif sbs_var in {"sigma-�00"}:
        kvar_name = "sgth0"
        kvar_units = "sigma-theta (p=0 db), kg/m^3-1000"
    elif sbs_var in {"sigma-t00"}:
        kvar_name = "sgt0"
        kvar_units = "sigma-t (p=0 db), kg/m^3-1000"
    elif sbs_var in {"sigma-100"}:
        kvar_name = "sg10"
        kvar_units = "sigma-1 (p=1000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-200"}:
        kvar_name = "sg20"
        kvar_units = "sigma-2 (p=2000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-300"}:
        kvar_name = "sg30"
        kvar_units = "sigma-3 (p=3000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-400"}:
        kvar_name = "sg40"
        kvar_units = "sigma-4 (p=4000 db), kg/m^3-1000"
    elif sbs_var in {"density11"}:
        kvar_name = "rho1"
        kvar_units = "density, kg/m^3"
    elif sbs_var in {"sigma-�11"}:
        kvar_name = "sgth1"
        kvar_units = "sigma-theta (p=0 db), kg/m^3-1000"
    elif sbs_var in {"sigma-t11"}:
        kvar_name = "sgt1"
        kvar_units = "sigma-t (p=0 db), kg/m^3-1000"
    elif sbs_var in {"sigma-111"}:
        kvar_name = "sg11"
        kvar_units = "sigma-1 (p=1000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-211"}:
        kvar_name = "sg21"
        kvar_units = "sigma-2 (p=2000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-311"}:
        kvar_name = "sg31"
        kvar_units = "sigma-3 (p=3000 db), kg/m^3-1000"
    elif sbs_var in {"sigma-411"}:
        kvar_name = "sg41"
        kvar_units = "sigma-4 (p=4000 db), kg/m^3-1000"

    # DENSITY DIFFERENCE
    elif sbs_var in {"D2-D1, d"}:
        kvar_name = "rhodiff"
        kvar_units = "density, kg/m^3"
    elif sbs_var in {"D2-D1,th"}:
        kvar_name = "sgthdiff"
        kvar_units = "sigma-theta, kg/m^3"
    elif sbs_var in {"D2-D1,t"}:
        kvar_name = "sgtdiff"
        kvar_units = "sigma-t, kg/m^3"
    elif sbs_var in {"D2-D1,1"}:
        kvar_name = "sg1diff"
        kvar_units = "sigma-1 (p=1000 db), kg/m^3"
    elif sbs_var in {"D2-D1,2"}:
        kvar_name = "sg2diff"
        kvar_units = "sigma-2 (p=1000 db), kg/m^3"
    elif sbs_var in {"D2-D1,3"}:
        kvar_name = "sg3diff"
        kvar_units = "sigma-3 (p=1000 db), kg/m^3"
    elif sbs_var in {"D2-D1,4"}:
        kvar_name = "sg4diff"
        kvar_units = "sigma-4 (p=1000 db), kg/m^3"

    # SOUND VELOCITY
    # primary sensor
    elif sbs_var in {"svCM"}:
        kvar_name = "sv0CM"
        kvar_units = "Chen-Millero, m/s"
    elif sbs_var in {"svCF"}:
        kvar_name = "sv0CF"
        kvar_units = "Chen-Millero, ft/s"
    elif sbs_var in {"svDM"}:
        kvar_name = "sv0DM"
        kvar_units = "Delgrosso, m/s"
    elif sbs_var in {"svDF"}:
        kvar_name = "sv0DF"
        kvar_units = "Delgrosso, ft/s"
    elif sbs_var in {"svWM"}:
        kvar_name = "sv0WM"
        kvar_units = "Wilson, m/s"
    elif sbs_var in {"svWF"}:
        kvar_name = "sv0WF"
        kvar_units = "Wilson, ft/s"
    # secondary sensor
    elif sbs_var in {"svCM1"}:
        kvar_name = "sv1CM"
        kvar_units = "Chen-Millero, m/s"
    elif sbs_var in {"svCF1"}:
        kvar_name = "sv1CF"
        kvar_units = "Chen-Millero, ft/s"
    elif sbs_var in {"svDM1"}:
        kvar_name = "sv1DM"
        kvar_units = "Delgrosso, m/s"
    elif sbs_var in {"svDF1"}:
        kvar_name = "sv1DF"
        kvar_units = "Delgrosso, ft/s"
    elif sbs_var in {"svWM1"}:
        kvar_name = "sv1WM"
        kvar_units = "Wilson, m/s"
    elif sbs_var in {"svWF1"}:
        kvar_name = "sv1WF"
        kvar_units = "Wilson, ft/s"
    # IOW sound velocity sensor
    elif sbs_var in {"iowSv"}:
        kvar_name = "iowSv"
        kvar_units = "IOW sound velocity sensor, m/s"
    elif sbs_var in {"sbeSv-iowSv"}:
        kvar_name = "svdiff"
        kvar_units = "SBE CTD - IOW SV sensor, m/s"
    # AVERAGE SOUND VELOCITY
    elif sbs_var in {"avgsvCM"}:
        kvar_name = "avgsvCM"
        kvar_units = "Chen-Millero, m/s"
    elif sbs_var in {"avgsvCF"}:
        kvar_name = "avgsvCF"
        kvar_units = "Chen-Millero, ft/s"
    elif sbs_var in {"avgsvDM"}:
        kvar_name = "avgsvDM"
        kvar_units = "Delgrosso, m/s"
    elif sbs_var in {"avgsvDF"}:
        kvar_name = "avgsvDF"
        kvar_units = "Delgrosso, ft/s"
    elif sbs_var in {"avgsvWM"}:
        kvar_name = "avgsvWM"
        kvar_units = "Wilson, m/s"
    elif sbs_var in {"avgsvWF"}:
        kvar_name = "avgsvWF"
        kvar_units = "Wilson, ft/s"

    # BUOYANCY
    elif sbs_var in {"N"}:
        kvar_name = "N"
        kvar_units = "cycles/hour"
    elif sbs_var in {"N^2"}:
        kvar_name = "N2"
        kvar_units = "rad^2/s^2"

    # ACCELERATION
    elif sbs_var in {"accM"}:
        kvar_name = "accM"
        kvar_units = "m/s^2"
    elif sbs_var in {"accF"}:
        kvar_name = "accF"
        kvar_units = "ft/s^2"

    # DESCENT RATE
    elif sbs_var in {"dz/dtM"}:
        kvar_name = "dzdtM"
        kvar_units = "m/s"
    elif sbs_var in {"dz/dtF"}:
        kvar_name = "dzdtF"
        kvar_units = "ft/s"

    # SEAFLOOR DEPTH
    elif sbs_var in {"sfdSM"}:
        kvar_name = "sfdSM"
        kvar_units = "salt water, m"
    elif sbs_var in {"sfdSF"}:
        kvar_name = "sfdSF"
        kvar_units = "salt water, ft"
    elif sbs_var in {"sfdFM"}:
        kvar_name = "sfdFM"
        kvar_units = "fresh water, m"
    elif sbs_var in {"sfdFF"}:
        kvar_name = "sfdFF"
        kvar_units = "fresh water, ft"

    # OTHER DERIVED VARIABLES
    # DYNAMIC METERS
    elif sbs_var in {"dm"}:
        kvar_name = "dm"
        kvar_units = "10 J/kg"
    # GEOPOTENTIAL ANOMALY
    elif sbs_var in {"gpa"}:
        kvar_name = "gpa"
        kvar_units = "J/kg"
    # PLUME ANOMALY
    elif sbs_var in {"pla"}:
        kvar_name = "pla"
        kvar_units = "?"
    # SPECIFIC VOLUME ANOMALY
    elif sbs_var in {"sva"}:
        kvar_name = "sva"
        kvar_units = "10^-8 *m^3/kg"
    # STABILITY
    elif sbs_var in {"E"}:
        kvar_name = "E"
        kvar_units = "rad^2/m"
    elif sbs_var in {"E10^-8"}:
        kvar_name = "E10e_8"
        kvar_units = "10^-8 *rad^2/m"
    # THERMOSTERIC ANOMALY
    elif sbs_var in {"tsa"}:
        kvar_name = "tsa"
        kvar_units = "10^-8 *m^3/kg"
    # SPECIFIC CONDUCTANCE
    elif sbs_var in {"specc"}:
        kvar_name = "specc"
        kvar_units = "uS/cm"
    elif sbs_var in {"speccumhoscm"}:
        kvar_name = "speccumhoscm"
        kvar_units = "umhos/cm"
    elif sbs_var in {"speccmsm"}:
        kvar_name = "speccmsm"
        kvar_units = "mS/cm"
    elif sbs_var in {"speccmmhoscm"}:
        kvar_name = "speccmmhoscm"
        kvar_units = "mmhos/cm"

    # BIOGEOCHEMICAL SENSORS

    # OXYGEN
    # RAW OXYGEN, SBE 43
    elif sbs_var in {"sbeox0V"}:
        kvar_name = "sbeox0V"
        kvar_units = "V"
    elif sbs_var in {"sbeox0F"}:
        kvar_name = "sbeox0F"
        kvar_units = "Hz"
    elif sbs_var in {"sbeox1V"}:
        kvar_name = "sbeox1V"
        kvar_units = "V"
    elif sbs_var in {"sbeox1F"}:
        kvar_name = "sbeox1F"
        kvar_units = "Hz"
    # DERIVED OXYGEN, SBE 43
    elif sbs_var in {"sbeox0ML/L"}:
        kvar_name = "sbeox0mL_L"
        kvar_units = "mL/L"
    elif sbs_var in {"sbeox0Mg/L"}:
        kvar_name = "sbeox0mg_L"
        kvar_units = "mg/L"
    elif sbs_var in {"sbeox0PS"}:
        kvar_name = "sbeox0PS"
        kvar_units = r"% saturation"
    elif sbs_var in {"sbeox0Mm/Kg"}:
        kvar_name = "sbeox0mum_kg"
        kvar_units = "micromol/kg"
    elif sbs_var in {"sbeox0Mm/L"}:
        kvar_name = "sbeox0mum_L"
        kvar_units = "micromole/L"
    elif sbs_var in {"sbeox0dOV/dT"}:
        kvar_name = "sbeox0dOV_dT"
        kvar_units = "V/s"
    elif sbs_var in {"sbeox1ML/L"}:
        kvar_name = "sbeox1mL_L"
        kvar_units = "mL/L"
    elif sbs_var in {"sbeox1Mg/L"}:
        kvar_name = "sbeox1mg_L"
        kvar_units = "mg/L"
    elif sbs_var in {"sbeox1PS"}:
        kvar_name = "sbeox1PS"
        kvar_units = r"% saturation"
    elif sbs_var in {"sbeox1Mm/Kg"}:
        kvar_name = "sbeox1mum_kg"
        kvar_units = "micromol/kg"
    elif sbs_var in {"sbeox1Mm/L"}:
        kvar_name = "sbeox1mum_L"
        kvar_units = "micromole/L"
    elif sbs_var in {"sbeox1dOV/dT"}:
        kvar_name = "sbeox1dOV_dT"
        kvar_units = "V/s"
    # Difference between SBE 43 sensors
    elif sbs_var in {"sbeox0ML/Ldiff"}:
        kvar_name = "sbeox0ML_Ldiff"
        kvar_units = "ml/l"
    elif sbs_var in {"sbeox0Mg/Ldiff"}:
        kvar_name = "sbeox0Mg_Ldiff"
        kvar_units = "mg/l"
    elif sbs_var in {"sbeox0PSdiff"}:
        kvar_name = "sbeox0PSdiff"
        kvar_units = r"% saturation"
    elif sbs_var in {"sbeox0Mm/Kgdiff"}:
        kvar_name = "sbeox0Mm_Kgdiff"
        kvar_units = "umol/kg"
    elif sbs_var in {"sbeox0Mm/Ldiff"}:
        kvar_name = "sbeox0Mm_Ldiff"
        kvar_units = "umol/l"
    # RAW OXYGEN, SBE 63
    elif sbs_var in {"sbeoxpd"}:
        kvar_name = "sbeoxpd"
        kvar_units = "usec"
    elif sbs_var in {"sbeoxpdv"}:
        kvar_name = "sbeoxpdv"
        kvar_units = "V"
    elif sbs_var in {"sbeoxpd1"}:
        kvar_name = "sbeoxpd1"
        kvar_units = "usec"
    elif sbs_var in {"sbeoxpdv1"}:
        kvar_name = "sbeoxpdv1"
        kvar_units = "sbeoxpdv1"
    elif sbs_var in {"sbeoxtv"}:
        kvar_name = "sbeoxtv"
        kvar_units = "sbeoxtv"
    elif sbs_var in {"sbeoxtv1"}:
        kvar_name = ""
        kvar_units = "sbeoxtv1"
    # RAW OXYGEN TEMPERATURE, SBE 63
    elif sbs_var in {"sbeoxTC"}:
        kvar_name = "sbeoxTC"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"sbeoxTF"}:
        kvar_name = "sbeoxTF"
        kvar_units = "ITS-90, deg F"
    elif sbs_var in {"sbeoxTC1"}:
        kvar_name = "sbeoxTC1"
        kvar_units = "ITS-90, deg C"
    elif sbs_var in {"sbeoxTF1"}:
        kvar_name = "sbeoxTF1"
        kvar_units = "ITS-90, deg F"
    # DERIVED OXYGEN, SBE 63
    elif sbs_var in {"sbeopoxML/L"}:
        kvar_name = "sbeopoxML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"sbeopoxMg/L"}:
        kvar_name = "sbeopoxMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"sbeopoxPS"}:
        kvar_name = "sbeopoxPS"
        kvar_units = r"% saturation"
    elif sbs_var in {"sbeopoxMm/Kg"}:
        kvar_name = "sbeopoxMm_Kg"
        kvar_units = "umol/kg"
    elif sbs_var in {"sbeopoxMm/L"}:
        kvar_name = "sbeopoxMm_L"
        kvar_units = "umol/l"
    elif sbs_var in {"sbeopoxML/L1"}:
        kvar_name = "sbeopoxML_L1"
        kvar_units = "ml/l"
    elif sbs_var in {"sbeopoxMg/L1"}:
        kvar_name = "sbeopoxMg_L1"
        kvar_units = "mg/l"
    elif sbs_var in {"sbeopoxPS1"}:
        kvar_name = "sbeopoxPS1"
        kvar_units = r"% saturation"
    elif sbs_var in {"SbeopoxMm/Kg1"}:
        kvar_name = "SbeopoxMm_Kg1"
        kvar_units = "umol/kg"
    elif sbs_var in {"sbeopoxMm/L1"}:
        kvar_name = "sbeopoxMm/L1"
        kvar_units = "umol/l"
    # AANDERAA OPTODE
    elif sbs_var in {"opoxML/L"}:
        kvar_name = "opoxML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"opoxMg/L"}:
        kvar_name = "opoxMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"opoxPS"}:
        kvar_name = "opoxPS"
        kvar_units = r"% saturation"
    elif sbs_var in {"opoxMm/L"}:
        kvar_name = "opoxMm_L"
        kvar_units = "umol/l"
    # BECKMAN/YSI
    elif sbs_var in {"oxC"}:
        kvar_name = "oxC"
        kvar_units = "uA"
    elif sbs_var in {"oxsC"}:
        kvar_name = "oxsC"
        kvar_units = "uA"
    elif sbs_var in {"oxTC"}:
        kvar_name = "oxTC"
        kvar_units = "deg C"
    elif sbs_var in {"oxTF"}:
        kvar_name = "oxTF"
        kvar_units = "deg F"
    elif sbs_var in {"oxsTC"}:
        kvar_name = "oxsTC"
        kvar_units = "deg C"
    elif sbs_var in {"oxsTF"}:
        kvar_name = "oxsTF"
        kvar_units = "deg F"
    elif sbs_var in {"oxML/L"}:
        kvar_name = "oxML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"oxMg/L"}:
        kvar_name = "oxMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"oxPS"}:
        kvar_name = "oxPS"
        kvar_units = r"% saturation"
    elif sbs_var in {"oxMm/Kg"}:
        kvar_name = "oxMm_Kg"
        kvar_units = "umol/kg"
    elif sbs_var in {"oxdOC/dT"}:
        kvar_name = "oxdOC_dT"
        kvar_units = "doc/dt"
    elif sbs_var in {"oxsML/L"}:
        kvar_name = "oxsML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"oxsMg/L"}:
        kvar_name = "oxsMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"oxsPS"}:
        kvar_name = "oxsPS"
        kvar_units = r"% saturation"
    elif sbs_var in {"oxsMm/Kg"}:
        kvar_name = "oxsMm_Kg"
        kvar_units = "umol/kg"
    elif sbs_var in {"oxsdOC/dT"}:
        kvar_name = "oxsdOC_dT"
        kvar_units = "doc/dt"
    # IOW OXYGEN SATURATIONS
    elif sbs_var in {"iowOxML/L"}:
        kvar_name = "iowOxML/L"
        kvar_units = "ml/l"
    # GARCIA & GORDON OXYGEN SATURATION
    elif sbs_var in {"oxsolML/L"}:
        kvar_name = "oxsolML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"oxsolMg/L"}:
        kvar_name = "oxsolMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"oxsolMm/Kg"}:
        kvar_name = "oxsolMm_Kg"
        kvar_units = "umol/kg"
    # WEISS OXYGEN SATURATION
    elif sbs_var in {"oxsatML/L"}:
        kvar_name = "oxsatML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"oxsatMg/L"}:
        kvar_name = "oxsatMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"oxsatMm/Kg"}:
        kvar_name = "oxsatMm_Kg"
        kvar_units = "umol/kg"

    # OPTICAL SENSORS
    # TURNER CYCLOPS
    # CDOM
    elif sbs_var in {"cdomflTC0", "cdomflTC1", "cdomflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb QS"
    # FLUORESCENCE
    elif sbs_var in {"chloroflTC0", "chloroflTC1", "chloroflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ug/l"
    # CRUDE OIL
    elif sbs_var in {"croilflTC0", "croilflTC1", "croilflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb QS"
    # FLUORESCEIN
    elif sbs_var in {"flflTC0", "flflTC1", "flflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb"
    # OPTICAL BRIGHTENERS
    elif sbs_var in {"obrflTC0", "obrflTC1", "obrflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb QS"
    # PHYCOCYANIN
    elif sbs_var in {"phycyflTC0", "phycyflTC1", "phycyflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "RFU"
    # PHYCOERYTHRIN
    elif sbs_var in {"phyeryflTC0", "phyeryflTC1", "phyeryflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "RFU"
    # REFINED FUELS
    elif sbs_var in {"rfuels0", "rfuels1", "rfuelsdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb NS"
    # RHODAMINE
    elif sbs_var in {"rhodflTC0", "rhodflTC1", "rhodflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ppb"
    # TURNER CYCLOPS TURBIDITY
    elif sbs_var in {"turbflTC0", "turbflTC1", "turbflTCdiff"}:
        kvar_name = sbs_var
        kvar_units = "NTU"

    # BIOSPHERICAL
    # FLUORESCENCE
    elif sbs_var in {"chConctr", "naFluor", "product"}:
        kvar_name = sbs_var
        kvar_units = ""
    # CHELSEA AQUA 3
    # FLUORESCENCE
    elif sbs_var in {"flC", "flC1", "flCdiff"}:
        kvar_name = sbs_var
        kvar_units = "ug/l"
    # CHELSEA MINI CHL CON
    # FLUORESCENCE
    elif sbs_var in {"flCM"}:
        kvar_name = "flCM"
        kvar_units = "ug/l"
    # CHELSEA UV AQUATRACKA
    # FLUORESCENCE
    elif sbs_var in {"flCUVA", "flCUVA1", "flCUVAdiff"}:
        kvar_name = sbs_var
        kvar_units = "ug/l"
    # DR HAARDT
    # Fluorescence: Chlorophyll a, Phycoerythrin, Yellow Sub
    elif sbs_var in {"haardtC", "haardtP", "haardtY"}:
        kvar_name = sbs_var
        kvar_units = ""
    # SEAPOINT
    # FLUORESCENCE
    elif sbs_var in {"flSP", "flSP1", "flSPdiff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # RHODAMINE
    elif sbs_var in {"flSPR"}:
        kvar_name = sbs_var
        kvar_units = ""
    # ULTRAVIOLET
    elif sbs_var in {"flSPuv0", "flSPuv1", "flSPuvdiff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # SEATECH FLUORESCENCE
    elif sbs_var in {"flS"}:
        kvar_name = sbs_var
        kvar_units = ""
    # TURNER 10-005
    elif sbs_var in {"flT"}:
        kvar_name = "flT"
        kvar_units = ""
    # TURNER 10-Au-005
    elif sbs_var in {"flTAu"}:
        kvar_name = "flTAu"
        kvar_units = ""
    # TURNER SCUFA CORRECTED
    elif sbs_var in {"flSCC", "flSCC1", "flSCCdiff"}:
        kvar_name = sbs_var
        kvar_units = "RFU"
    # TURNER SCUFA
    elif sbs_var in {"flScufa", "flScufa1", "flScufadiff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # WETLABS AC3
    elif sbs_var in {"wetChAbs"}:
        kvar_name = "wetChAbs"
        kvar_units = "1/m"
    # WETLABS CDOM
    elif sbs_var in {
        "wetCDOM",
        "wetCDOM1",
        "wetCDOM2",
        "wetCDOM3",
        "wetCDOM4",
        "wetCDOM5",
        "wetCDOMdiff",
    }:
        kvar_name = sbs_var
        kvar_units = "mg/m^3"
    # WETLABS AC3 CHLOROPHYLL
    elif sbs_var in {"wetChConc"}:
        kvar_name = "wetChConc"
        kvar_units = "mg/m^3"
    # WETLABS ECO-AFL
    elif sbs_var in {"flECO-AFL"}:
        kvar_name = "flECO_AFL"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL1"}:
        kvar_name = "flECO_AFL1"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL2"}:
        kvar_name = "flECO_AFL2"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL3"}:
        kvar_name = "flECO_AFL3"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL4"}:
        kvar_name = "flECO_AFL4"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFL5"}:
        kvar_name = "flECO_AFL5"
        kvar_units = "mg/m^3"
    elif sbs_var in {"flECO-AFLdiff"}:
        kvar_name = "flECO_AFLdiff"
        kvar_units = "mg/m^3"
    # WETLABS SEAOWL
    # FLUOROMETER
    elif sbs_var in {"flWETSeaOWLchl0", "flWETSeaOWLchl1", "flWETSeaOWLchldiff"}:
        kvar_name = sbs_var
        kvar_units = "?g/l"
    # FDOM
    elif sbs_var in {"flWETSeaOWLfdom0", "flWETSeaOWLfdom1", "flWETSeaOWLfdomdiff"}:
        kvar_name = sbs_var
        kvar_units = "?g/l"
    # WETSTAR
    elif sbs_var in {
        "wetStar",
        "wetStar1",
        "wetStar2",
        "wetStar3",
        "wetStar4",
        "wetStar5",
        "wetStardiff",
    }:
        kvar_name = sbs_var
        kvar_units = "mg/m^3"

    # BACKSCATTER
    # D & A BACKSCATTERANCE
    elif sbs_var in {"obs", "obs1", "obsdiff"}:
        kvar_name = sbs_var
        kvar_units = "NTU"
    # CHELSEA NEPHELOMETER
    elif sbs_var in {"nephc"}:
        kvar_name = "nephc"
        kvar_units = "FTU"
    # D & A
    elif sbs_var in {"obs3+"}:
        kvar_name = "obs3plus"
        kvar_units = "NTU"
    elif sbs_var in {"obs3+1"}:
        kvar_name = "obs3plus1"
        kvar_units = "NTU"
    elif sbs_var in {"obs3+diff"}:
        kvar_name = "obs3plusdiff"
        kvar_units = "NTU"
    # DR. HAARDT
    elif sbs_var in {"haardtT"}:
        kvar_name = "haardtT"
        kvar_units = ""
    # IFREMER
    elif sbs_var in {"diff"}:
        kvar_name = "diff"
        kvar_units = ""
    # SEATECH LS6000
    elif sbs_var in {"stLs6000", "stLs60001", "stLs6000diff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # TURNER SCUFA
    elif sbs_var in {"obsscufa", "obsscufa1", "obsscufadiff"}:
        kvar_name = sbs_var
        kvar_units = ""
    # WETLABS RAW COUNTS
    elif sbs_var in {"wl0", "wl1", "wl2", "wl3", "wl4", "wl5"}:
        kvar_name = sbs_var
        kvar_units = "Counts"

    # TURBIDITY
    # SEAPOINT TURBIDITY
    elif sbs_var in {"seaTurbMtr", "seaTurbMtr1", "seaTurbMtrdiff"}:
        kvar_name = sbs_var
        kvar_units = "FTU"
    # WETLABS ECO BB
    elif sbs_var in {
        "turbWETbb0",
        "turbWETbb1",
        "turbWETbb2",
        "turbWETbb3",
        "turbWETbb4",
        "turbWETbb5",
        "turbWETbbdiff",
    }:
        kvar_name = sbs_var
        kvar_units = "m^-1/sr"
    # WETLABS ECO
    elif sbs_var in {
        "turbWETntu0",
        "turbWETntu1",
        "turbWETntu2",
        "turbWETntu3",
        "turbWETntu4",
        "turbWETntu5",
        "turbWETntudiff",
    }:
        kvar_name = sbs_var
        kvar_units = "NTU"
    # WETLABS SEAOWL
    elif sbs_var in {"turbWETSeaOWLbb0", "turbWETSeaOWLbb1", "turbWETSeaOWLbbdiff"}:
        kvar_name = sbs_var
        kvar_units = "m^-1/sr"

    # TRANSMISSOMETERS
    # CHELSEA/SEATECH BEAM ATTENUATION
    elif sbs_var in {"bat", "bat1", "batdiff"}:
        kvar_name = sbs_var
        kvar_units = "1/m"
    # WETLABS AC3 BEAM TRANSMISSION
    elif sbs_var in {"wetBAttn"}:
        kvar_name = "wetBAttn"
        kvar_units = "1/m"
    # WETLABS C-STAR BEAM ATTENUATION
    elif sbs_var in {
        "CStarAt",
        "CStarAt0",
        "CStarAt1",
        "CStarAt2",
        "CStarAt3",
        "CStarAt4",
        "CStarAt5",
        "CStarAtdiff",
    }:
        kvar_name = sbs_var
        kvar_units = "1/m"
    # CHELSEA/SEATECH BEAM TRANSMISSION
    elif sbs_var in {"xmiss", "xmiss1", "xmissdiff"}:
        kvar_name = sbs_var
        kvar_units = r"%"
    # WETLABS AC3 BEAM TRANSMISSION
    elif sbs_var in {"wetBTrans"}:
        kvar_name = "wetBTrans"
        kvar_units = r"%"
    # WETLABS C-STAR BEAM TRANSMISSION
    elif sbs_var in {
        "CStarTr",
        "CStarTr0",
        "CStarTr1",
        "CStarTr2",
        "CStarTr3",
        "CStarTr4",
        "CStarTr5",
        "CStarTrdiff",
    }:
        kvar_name = sbs_var
        kvar_units = r"%"

    # LISST
    elif sbs_var in {"lisstBC"}:
        kvar_name = "lisstBC"
        kvar_units = "1/m"
    elif sbs_var in {"lisstOT"}:
        kvar_name = "lisstOT"
        kvar_units = r"%"
    elif sbs_var in {"lisstMD"}:
        kvar_name = "lisstMD"
        kvar_units = "u"
    elif sbs_var in {"lisstTVC"}:
        kvar_name = "lisstTVC"
        kvar_units = "ul/l"

    # PAR
    elif sbs_var in {"cpar"}:
        kvar_name = "cpar"
        kvar_units = r"%"
    elif sbs_var in {"par"}:
        kvar_name = "par"
        kvar_units = ""
    elif sbs_var in {"par1"}:
        kvar_name = "par1"
        kvar_units = ""
    elif sbs_var in {"par/log"}:
        kvar_name = "par_log"
        kvar_units = "umol photons/m2/s"
    elif sbs_var in {"spar"}:
        kvar_name = "spar"
        kvar_units = ""

    # BIOGEOCHEMICAL SENSORS
    # METHANE FRANATECH METS
    elif sbs_var in {"meth"}:
        kvar_name = "meth"
        kvar_units = "umol/l"
    elif sbs_var in {"methT"}:
        kvar_name = "methT"
        kvar_units = "deg C"
    # NITROGEN
    elif sbs_var in {"n2satML/L"}:
        kvar_name = "n2satML_L"
        kvar_units = "ml/l"
    elif sbs_var in {"n2satMg/L"}:
        kvar_name = "n2satMg_L"
        kvar_units = "mg/l"
    elif sbs_var in {"n2satumol/kg"}:
        kvar_name = "n2satumol_kg"
        kvar_units = "umol/kg"
    # OXIDATION REDUCTION POTENTIAL
    elif sbs_var in {"orp"}:
        kvar_name = "orp"
        kvar_units = "mV"
    # pH
    elif sbs_var in {"ph", "phInt", "phExt"}:
        kvar_name = sbs_var
        kvar_units = ""
    # ZAPS
    elif sbs_var in {"zaps"}:
        kvar_name = "zaps"
        kvar_units = "nmol"

    # USER DEFINED VARIABLE
    elif sbs_var in {"user", "user1", "user2", "user3", "user4", "user5"}:
        kvar_name = sbs_var
        kvar_units = ""
    # USER POLYNOMIAL
    elif sbs_var in {"upoly0", "upoly1", "upoly2"}:
        kvar_name = sbs_var
        kvar_units = ""

    # PACKAGE STATUS
    # ALTIMETER
    elif sbs_var in {"altM"}:
        kvar_name = "altM"
        kvar_units = "m"
    elif sbs_var in {"altF"}:
        kvar_name = "altF"
        kvar_units = "ft"

        # Bottom Contact
    elif sbs_var in {"bct"}:
        kvar_name = "bct"
        kvar_units = ""

    # WATER SAMPLER STATUS
    # BOTTLE POSITION, BOTTLES CLOSED (HB), BOTTLES FIRED, NEW POSITION'
    elif sbs_var in {"bpos", "HBBotCls", "nbf", "newpos"}:
        kvar_name = sbs_var
        kvar_units = ""
    # PUMP STATUS
    elif sbs_var in {"pumps"}:
        kvar_name = "pumps"
        kvar_units = ""

    # DATA STREAM
    # BYTE COUNT
    elif sbs_var in {"nbytes"}:
        kvar_name = "nbytes"
        kvar_units = ""
    # MODULO ERROR COUNT
    elif sbs_var in {"modError"}:
        kvar_name = "modError"
        kvar_units = ""
    # MODULO WORD
    elif sbs_var in {"mod"}:
        kvar_name = "mod"
        kvar_format = r"%s"
        kvar_units = ""
    # SCAN COUNT
    elif sbs_var in {"scan"}:
        kvar_name = "scan"
        kvar_units = ""
    # SCANS PER BIN
    elif sbs_var in {"nbin"}:
        kvar_name = "nbin"
        kvar_units = ""

    # FLAGS
    elif sbs_var in {"flag"}:
        kvar_name = "flag"
        kvar_units = ""

    else:
        # assign the original SBS name to the output variable
        kvar_name = sbs_var
        # find and replace all the characters that are not alphabetic,
        # numbers or underscores with underscores
        kvar_name = re.sub(r"\W", "_", kvar_name)
        # print(f"Assigning {sbs_var} as {kvar_name}")
        kvar_units = ""

    #     elif sbs_var in  ''
    #         kvar_name = ''
    # #         kvar_units = ''

    return {"name": kvar_name, "format": kvar_format, "units": kvar_units}



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""A collection of raw data conversion functions.

Functions:

    convert_temperature (np.ndarray, TemperatureCoefficients, str, str, bool)
    convert_pressure ( np.ndarray, np.ndarray, PressureCoefficients, str)
    convert_conductivity (np.ndarray, np.ndarray, np.ndarray, ConductivityCoefficients)
    potential_density_from_t_s_p (np.ndarray, np.ndarray, np.ndarray, float, float, float)
    potential_density_from_t_c_p (np.ndarray, np.ndarray, np.ndarray, float, float, float)
    density_from_t_s_p (np.ndarray, np.ndarray, np.ndarray, float, float)
    density_from_t_c_p (np.ndarray, np.ndarray, np.ndarray, float, float)
    depth_from_pressure (np.ndarray, float, depth_units="m", pressure_units="dbar")
    convert_sbe63_oxygen (np.ndarray, np.ndarray, np.ndarray, np.ndarray, Oxygen63Coefficients, Thermistor63Coefficients, str)
    convert_sbe63_thermistor (np.ndarray, Thermistor63Coefficients)
    convert_sbe43_oxygen (np.ndarray, np.ndarray, np.ndarray, np.ndarray, Oxygen43Coefficients, bool, bool, float, float)
    convert_oxygen_to_mg_per_l (np.ndarray)
    convert_oxygen_to_umol_per_kg (np.ndarray, np.ndarray)
    convert_eco_chlorophylla_val (float, ChlorophyllACoefficients)
    convert_eco_turbidity_val (float, TurbidityCoefficients)
    convert_sbe18_ph_val(float, float, PH18Coefficients)
    convert_par_logarithmic_val(float, PARCoefficients)

"""

# Native imports
from math import sqrt, e, log, exp, floor
from typing import Literal

# Third-party imports
import gsw
import numpy as np
from scipy import stats

# Sea-Bird imports

# Internal imports
from .cal_coefficients import (
    ConductivityCoefficients,
    Oxygen43Coefficients,
    Oxygen63Coefficients,
    PARCoefficients,
    PH18Coefficients,
    PHSeaFETInternalCoefficients,
    PHSeaFETExternalCoefficients,
    PressureCoefficients,
    TemperatureCoefficients,
    Thermistor63Coefficients,
    ECOCoefficients,
)


DBAR_TO_PSI = 1.450377
PSI_TO_DBAR = 0.6894759
OXYGEN_PHASE_TO_VOLTS = 39.457071
KELVIN_OFFSET_0C = 273.15
KELVIN_OFFSET_25C = 298.15
OXYGEN_MLPERL_TO_MGPERL = 1.42903
OXYGEN_MLPERL_TO_UMOLPERKG = 44660
ITS90_TO_IPTS68 = 1.00024  # taken from https://blog.seabird.com/ufaqs/what-is-the-difference-in-temperature-expressions-between-ipts-68-and-its-90/
UMNO3_TO_MGNL = 0.014007
R = 8.3144621  # [J K^{-1} mol^{-1}] Gas constant, NIST Reference on Constants retrieved 10-05-2015
F = 96485.365  # [Coulombs mol^{-1}] Faraday constant, NIST Reference on Constants retrieved 10-05-2015


def convert_temperature(
    temperature_counts_in: np.ndarray,
    coefs: TemperatureCoefficients,
    standard: Literal["ITS90", "IPTS68"] = "ITS90",
    units: Literal["C", "F"] = "C",
    use_mv_r: bool = False,
):
    """Returns the value after converting it to degrees C, ITS-90.

    Data is expected to be raw data from an instrument in A/D counts

    Args:
        temperature_counts_in (np.ndarray): temperature value to convert in A/D counts
        coefs (TemperatureCoefficients) calibration coefficients for the temperature sensor
        standard (str): whether to use ITS90 or to use IPTS-68 calibration standard
        units (str): whether to use celsius or to convert to fahrenheit
        use_mv_r (bool): true to perform extra conversion steps required
        by some instruments (check the cal sheet to see if this is required)

    Returns:
        int: temperature val converted to ITS-90 degrees C
    """

    if use_mv_r:
        mv = (temperature_counts_in - 524288) / 1.6e007
        r = (mv * 2.900e009 + 1.024e008) / (2.048e004 - mv * 2.0e005)
        temperature_counts = r
    else:
        temperature_counts = temperature_counts_in

    log_t = np.log(temperature_counts)
    temperature = (
        1 / (coefs.a0 + coefs.a1 * log_t + coefs.a2 * log_t**2 + coefs.a3 * log_t**3)
    ) - KELVIN_OFFSET_0C

    if standard == "IPTS68":
        temperature *= ITS90_TO_IPTS68
    if units == "F":
        temperature = temperature * 9 / 5 + 32  # Convert C to F

    return temperature


def convert_pressure(
    pressure_count: np.ndarray,
    compensation_voltage: np.ndarray,
    coefs: PressureCoefficients,
    units: Literal["dbar", "psia"] = "psia",
):
    """Returns the value after converting it to psia (pounds per square inch, abolute).

    pressure_count and compensation_voltage are expected to be raw data from an instrument in A/D counts

    Args:
        pressure_count (np.ndarray): pressure value to convert, in A/D counts
        compensation_voltage (np.ndarray): pressure temperature compensation voltage,
            in counts or volts depending on the instrument
        coefs (PressureCoefficients): calibration coefficients for the pressure sensor
        units (str): whether or not to use psia or dbar as the returned unit type

    Returns:
        int: pressure val in PSIA
    """
    sea_level_pressure = 14.7

    t = (
        coefs.ptempa0
        + coefs.ptempa1 * compensation_voltage
        + coefs.ptempa2 * compensation_voltage**2
    )
    x = pressure_count - coefs.ptca0 - coefs.ptca1 * t - coefs.ptca2 * t**2
    n = x * coefs.ptcb0 / (coefs.ptcb0 + coefs.ptcb1 * t + coefs.ptcb2 * t**2)
    pressure = coefs.pa0 + coefs.pa1 * n + coefs.pa2 * n**2 - sea_level_pressure

    if units == "dbar":
        pressure *= PSI_TO_DBAR

    return pressure


def convert_conductivity(
    conductivity_count: np.ndarray,
    temperature: np.ndarray,
    pressure: np.ndarray,
    coefs: ConductivityCoefficients,
):
    """Returns the value after converting it to S/m.

    Data is expected to be raw data from instrument in A/D counts

    Args:
        conductivity_count (np.ndarray): conductivity value to convert, in A/D counts
        temperature (np.ndarray): temperature value to use are reference, in degrees C
        pressure (np.ndarray): pressure value to use are reference, in dbar
        coefs (float): coefs calibration coefficient for the conductivity sensor
    Returns:
        Decimal: conductivity val converted to S/m
    """
    f = conductivity_count * np.sqrt(1 + coefs.wbotc * temperature) / 1000
    numerator = coefs.g + coefs.h * f**2 + coefs.i * f**3 + coefs.j * f**4
    denominator = 1 + coefs.ctcor * temperature + coefs.cpcor * pressure
    return numerator / denominator


def potential_density_from_t_s_p(
    temperature: np.ndarray,
    salinity: np.ndarray,
    pressure: np.ndarray,
    lon=0.0,
    lat=0.0,
    reference_pressure=0.0,
):
    """Derive potential density from measured temperature, salinity, and pressure.

    See: TEOS_10.cpp line 953

    Args:
        temperature (np.ndarray): Measure temperature in degrees C
        salinity (np.ndarray): Measured salinity in practical salinity units
        pressure (np.ndarray): Measured pressure in decibars
        lon (float): Longitude
        lat (float): Latitude
        reference_pressure (float, optional): Reference pressure in decibars. Defaults to 0.0.

    Returns:
        np.ndarray: Potential density in kg/m^3
    """

    absolute_salinity = gsw.SA_from_SP(salinity, pressure, lon, lat)
    conservative_temperature = gsw.CT_from_t(absolute_salinity, temperature, pressure)
    potential_density = (
        gsw.rho(absolute_salinity, conservative_temperature, reference_pressure) - 1000
    )
    return potential_density


def potential_density_from_t_c_p(
    temperature: np.ndarray,
    conductivity: np.ndarray,
    pressure: np.ndarray,
    lon=0.0,
    lat=0.0,
    reference_pressure=0.0,
):
    """Derive potential density from measured temperature, salinity, and pressure.

    See: TEOS_10.cpp line 953

    Args:
        temperature (np.ndarray): Measure temperature in degrees C
        conductivity (np.ndarray): Measured conductivity in mSiemens/cm
        pressure (np.ndarray): Measured pressure in decibars
        lon (float): Longitude
        lat (float): Latitude
        reference_pressure (float, optional): Reference pressure in decibars. Defaults to 0.0.

    Returns:
        np.ndarray: Potential density in kg/m^3
    """

    salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    return potential_density_from_t_s_p(
        temperature, salinity, pressure, lon, lat, reference_pressure
    )


def density_from_t_s_p(
    temperature: np.ndarray,
    salinity: np.ndarray,
    pressure: np.ndarray,
    lon=0.0,
    lat=0.0,
):
    """Derive potential density from measured temperature, salinity, and pressure.

    See: TEOS_10.cpp line 953

    Args:
        temperature (np.ndarray): Measure temperature in degrees C
        salinity (np.ndarray): Measured salinity in practical salinity units
        pressure (np.ndarray): Measured pressure in decibars
        lon (float): Longitude
        lat (float): Latitude

    Returns:
        np.ndarray: Potential density in kg/m^3
    """

    absolute_salinity = gsw.SA_from_SP(salinity, pressure, lon, lat)
    conservative_temperature = gsw.CT_from_t(absolute_salinity, temperature, pressure)
    density = gsw.rho(absolute_salinity, conservative_temperature, pressure)
    return density


def density_from_t_c_p(
    temperature: np.ndarray,
    conductivity: np.ndarray,
    pressure: np.ndarray,
    lon=0.0,
    lat=0.0,
):
    """Derive potential density from measured temperature, salinity, and pressure.

    See: TEOS_10.cpp line 953

    Args:
        temperature (np.ndarray): Measure temperature in degrees C
        conductivity (np.ndarray): Measured conductivity in mSiemens/cm
        pressure (np.ndarray): Measured pressure in decibars
        lon (float): Longitude
        lat (float): Latitude

    Returns:
        np.ndarray: Potential density in kg/m^3
    """

    salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    return density_from_t_s_p(temperature, salinity, pressure, lon, lat)


def depth_from_pressure(
    pressure_in: np.ndarray,
    latitude: float,
    depth_units: Literal["m", "ft"] = "m",
    pressure_units: Literal["dbar", "psi"] = "dbar",
):
    """Derive depth from pressure and latitude.

    Args:
        pressure (np.ndarray): Numpy array of floats representing pressure in dbar or psi
        latitude (float): Latitude (-90.0 to 90.0)
        depth_units (str, optional): 'm' for meters, 'ft' for feet. Defaults to 'm'.
        pressure_units (str, optional): 'dbar' for decibars, 'psi' for PSI. Defaults to 'dbar'.

    Returns:
        np.ndarray: A numpy array representing depth in meters or feet
    """
    pressure = pressure_in.copy()
    if pressure_units == "psi":
        pressure /= DBAR_TO_PSI

    depth = -gsw.z_from_p(pressure, latitude)

    if depth_units == "ft":
        depth *= 3.28084

    return depth


def convert_sbe63_oxygen(
    raw_oxygen_phase: np.ndarray,
    thermistor: np.ndarray,
    pressure: np.ndarray,
    salinity: np.ndarray,
    coefs: Oxygen63Coefficients,
    thermistor_coefs: Thermistor63Coefficients,
    thermistor_units: Literal["volts", "C"] = "volts",
):
    """Returns the data after converting it to ml/l.

        raw_oxygen_phase is expected to be in raw phase, raw_thermistor_temp in counts, pressure in dbar, and salinity in practical salinity (PSU)

        Args:
            raw_oxygen_phase (np.ndarray): SBE63 phase value, in microseconds
    #         raw_thermistor_temp (np.ndarray): SBE63 thermistor data to use are reference, in counts
            pressure (np.ndarray): Converted pressure value from the attached CTD, in dbar
            salinity (np.ndarray): Converted salinity value from the attached CTD, in practical salinity PSU
            coefs (Oxygen63Coefficients): calibration coefficients for the SBE63 sensor
        Returns:
            np.ndarray: converted Oxygen value, in ml/l
    """
    if thermistor_units == "volts":
        temperature = convert_sbe63_thermistor(thermistor, thermistor_coefs)
    elif thermistor_units == "C":
        temperature = thermistor

    oxygen_volts = raw_oxygen_phase / OXYGEN_PHASE_TO_VOLTS  # from the manual
    # O2 (ml/L) = [((a0 + a1T + a2(V^2)) / (b0 + b1V) – 1) / Ksv] [SCorr] [PCorr]

    # Ksv = c0 + c1T + c2 (T^2)
    ksv = coefs.c0 + coefs.c1 * temperature + coefs.c2 * temperature**2

    # SCorr = exp [S * (SolB0 + SolB1 * Ts + SolB2 * Ts^2 + SolB3 * Ts^3) + SolC0 * S^2]
    # The following correction coefficients are all constants
    SOL_B0 = -6.24523e-3
    SOL_B1 = -7.37614e-3
    SOL_B2 = -1.0341e-2
    SOL_B3 = -8.17083e-3
    SOL_C0 = -4.88682e-7

    # Ts = ln [(298.15 – T) / (273.15 + T)]
    ts = np.log((KELVIN_OFFSET_25C - temperature) / (KELVIN_OFFSET_0C + temperature))
    s_corr_exp = (
        salinity * (SOL_B0 + SOL_B1 * ts + SOL_B2 * ts**2 + SOL_B3 * ts**3) + SOL_C0 * salinity**2
    )
    s_corr = e**s_corr_exp

    # Pcorr = exp (E * P / K)
    K = temperature + KELVIN_OFFSET_0C
    # temperature in Kelvin
    p_corr_exp = (e * pressure) / K
    p_corr = e**p_corr_exp

    ox_val = (
        (
            (
                (coefs.a0 + coefs.a1 * temperature + coefs.a2 * oxygen_volts**2)
                / (coefs.b0 + coefs.b1 * oxygen_volts)
                - 1.0
            )
            / ksv
        )
        * s_corr
        * p_corr
    )

    return ox_val


def convert_sbe63_thermistor(
    instrument_output: np.ndarray,
    coefs: Thermistor63Coefficients,
):
    """Converts a SBE63 thermistor raw output array to temperature in ITS-90 deg C.

    Args:
        instrument_output (np.ndarray) raw values from the thermistor
        coefs (Thermisto63Coefficients): calibration coefficients for the thermistor in the SBE63 sensor

    Returns:
        np.ndarray: converted thermistor temperature values in ITS-90 deg C
    """
    logVal = np.log((100000 * instrument_output) / (3.3 - instrument_output))
    temperature = (
        1 / (coefs.ta0 + coefs.ta1 * logVal + coefs.ta2 * logVal**2 + coefs.ta3 * logVal**3)
        - KELVIN_OFFSET_0C
    )
    return temperature


def convert_sbe43_oxygen(
    voltage: np.ndarray,
    temperature: np.ndarray,
    pressure: np.ndarray,
    salinity: np.ndarray,
    coefs: Oxygen43Coefficients,
    apply_tau_correction: bool = False,
    apply_hysteresis_correction: bool = False,
    window_size: float = 1,
    sample_interval: float = 1,
):
    """Returns the data after converting it to ml/l.

    voltage is expected to be in volts, temperature in deg c, pressure in dbar, and salinity in practical salinity (PSU)
    All equation information comes from the June 2013 revision of the SBE43 manual

    Args:
        voltage (float): SBE43 voltage
        temperature (float): temperature value converted to deg C
        pressure (float): Converted pressure value from the attached CTD, in dbar
        salinity (float): Converted salinity value from the attached CTD, in practical salinity PSU
        coefs (Oxygen43Coefficients): calibration coefficients for the SBE43 sensor
        apply_tau_correction (bool): whether or not to run tau correction
        apply_hysteresis_correction (bool): whether or not to run hysteresis correction
        window_size (float): size of the window to use for tau correction, if applicable. In seconds.
        sample_interval (float): sample rate of the data to be used for tau correction, if applicable. In seconds.

    Returns:
        np.ndarray: converted Oxygen values, in ml/l
    """
    # start with all 0 for the dvdt
    dvdt_values = np.zeros(len(voltage))
    if apply_tau_correction:
        # Calculates how many scans to have on either side of our median point, accounting for going out of index bounds
        scans_per_side = floor(window_size / 2 / sample_interval)
        for i in range(scans_per_side, len(voltage) - scans_per_side):
            ox_subset = voltage[i - scans_per_side : i + scans_per_side + 1]

            time_subset = np.arange(
                0, len(ox_subset) * sample_interval, sample_interval, dtype=float
            )

            result = stats.linregress(time_subset, ox_subset)

            dvdt_values[i] = result.slope

    correct_ox_voltages = voltage.copy()
    if apply_hysteresis_correction:
        # Hysteresis starts at 1 because 0 can't be corrected
        for i in range(1, len(correct_ox_voltages)):
            # All Equation info from APPLICATION NOTE NO. 64-3
            d = 1 + coefs.h1 * (np.exp(pressure[i] / coefs.h2) - 1)
            c = np.exp(-1 * sample_interval / coefs.h3)
            ox_volts = correct_ox_voltages[i] + coefs.v_offset

            prev_ox_volts_new = correct_ox_voltages[i - 1] + coefs.v_offset
            ox_volts_new = ((ox_volts + prev_ox_volts_new * c * d) - (prev_ox_volts_new * c)) / d
            ox_volts_final = ox_volts_new - coefs.v_offset
            correct_ox_voltages[i] = ox_volts_final

    oxygen = _convert_sbe43_oxygen(
        correct_ox_voltages,
        temperature,
        pressure,
        salinity,
        coefs,
        dvdt_values,
    )
    return oxygen


def _convert_sbe43_oxygen(
    voltage: np.ndarray,
    temperature: np.ndarray,
    pressure: np.ndarray,
    salinity: np.ndarray,
    coefs: Oxygen43Coefficients,
    dvdt_value: np.ndarray,
):
    """Returns the data after converting it to ml/l.

    voltage is expected to be in volts, temperature in deg c, pressure in dbar, and salinity in practical salinity (PSU)
    All equation information comes from the June 2013 revision of the SBE43 manual.
    Expects that hysteresis correction is already performed on the incoming voltage, if desired.

    Args:
        voltage (float): SBE43 voltage
        temperature (float): temperature value converted to deg C
        pressure (float): Converted pressure value from the attached CTD, in dbar
        salinity (float): Converted salinity value from the attached CTD, in practical salinity PSU
        coefs (Oxygen43Coefficients): calibration coefficients for the SBE43 sensor
        dvdt_value (float): derivative value of voltage with respect to time at this point. Expected to be 0 if not using Tau correction

    Returns:
        float: converted Oxygen value, in ml/l
    """

    # Oxygen Solubility equation constants, From SBE43 Manual Appendix A
    A0 = 2.00907
    A1 = 3.22014
    A2 = 4.0501
    A3 = 4.94457
    A4 = -0.256847
    A5 = 3.88767
    B0 = -0.00624523
    B1 = -0.00737614
    B2 = -0.010341
    B3 = -0.00817083
    C0 = -0.000000488682

    ts = np.log((KELVIN_OFFSET_25C - temperature) / (KELVIN_OFFSET_0C + temperature))
    a_term = A0 + A1 * ts + A2 * ts**2 + A3 * ts**3 + A4 * ts**4 + A5 * ts**5
    b_term = salinity * (B0 + B1 * ts + B2 * ts**2 + B3 * ts**3)
    c_term = C0 * salinity**2
    solubility = np.exp(a_term + b_term + c_term)

    # Tau correction
    tau = coefs.tau_20 * np.exp(coefs.d1 * pressure + coefs.d2 * (temperature - 20)) * dvdt_value

    soc_term = coefs.soc * (voltage + coefs.v_offset + tau)
    temp_term = 1.0 + coefs.a * temperature + coefs.b * temperature**2 + coefs.c * temperature**3
    oxygen = (
        soc_term
        * solubility
        * temp_term
        * np.exp((coefs.e * pressure) / (temperature + KELVIN_OFFSET_0C))
    )
    return oxygen


def convert_oxygen_to_mg_per_l(ox_values: np.ndarray):
    """Converts given oxygen values to milligrams/Liter.

    Expects oxygen values to be in Ml/L

    Args:
        ox_values (np.ndarray): oxygen values, already converted to ml/L

    Returns:
        np.ndarray: oxygen values converted to milligrams/Liter
    """
    # From the SBE43 and SBE63 manual
    return ox_values * OXYGEN_MLPERL_TO_MGPERL


def convert_oxygen_to_umol_per_kg(ox_values: np.ndarray, potential_density: np.ndarray):
    """Converts given oxygen values to milligrams/Liter.

    Expects oxygen values to be in Ml/L
    Note: Sigma-Theta is expected to be calculated via gsw_sigma0, meaning is it technically potential density anomaly.
    Calculating using gsw_rho(SA, CT, p_ref = 0) results in actual potential density, but this function already does the converison,
    So values will need to have 1000 subtracted from them before being passed into this function.
    The function is done this way to stay matching to the manual for the SBE63 and SBE43, but the results of either method are identical.

    Args:
        ox_values (np.ndarray): oxygen values, already converted to ml/L
        potential_density (np.ndarray): potential density (sigma-theta) values.
                                        Expected to be the same length as ox_values

    Returns:
        np.ndarray: oxygen values converted to milligrams/Liter
    """
    # From the SBE43 and SBE63 manual
    convertedVals = np.divide(ox_values * OXYGEN_MLPERL_TO_UMOLPERKG, (potential_density + 1000))
    return convertedVals


def convert_eco(
    raw: float,
    coefs: ECOCoefficients,
):
    """Converts a raw value for any ECO measurand.

    Args:
        raw (float): raw counts for digital, raw volts for analog
        coefs (ChlorophyllACoefficients): calibration coefficients

    Returns:
        float: converted ECO measurement in calibration units
    """
    converted = coefs.slope * (raw - coefs.offset)

    return converted


def convert_sbe18_ph(
    raw_ph: float,
    temperature: float,
    coefs: PH18Coefficients,
):
    """Converts a raw voltage value for pH
        All equation information comes from application note 18-1
    Args:
        raw_ph (float): raw output voltage from pH sensor (0-5V)
        temperature (float): temperature value to use for temperature compensation in degrees C
        coefs (PH18Coefficients): slope and offset for the pH sensor
    Returns:
        float: converted pH
    """
    pH = 7 + (raw_ph - coefs.offset) / (
        1.98416e-4 * (temperature + KELVIN_OFFSET_0C) * coefs.slope
    )
    return pH


def convert_par_logarithmic(
    raw_par: float,
    coefs: PARCoefficients,
):
    """Converts a raw voltage value for PAR to µmol photons/m2*s
        All equation information comes from application note 96
    Args:
        rawpH (float): raw output voltage from PAR sensor
        coefs (PARCoefficients): calibration coefficients for the PAR sensor
    Returns:
        float: converted PAR in µmol photons/m2*s
    """
    PAR = coefs.multiplier * coefs.im * 10 ** ((raw_par - coefs.a0) / coefs.a1)

    return PAR


def convert_nitrate(
    volts: np.ndarray, dac_min: float, dac_max: float, units: Literal["uMNO3", "mgNL"] = "uMNO3"
):
    """Convert SUNA raw voltages to uMNO3 or mgNL

    :param volts: raw output voltage from a SUNA
    :param dac_min: NO3 value that corresponds to V_MIN
    :param dac_max: NO3 value that corresponds to V_MAX
    :param units: conversion output units, defaults to 'uMNO3'
    :return: converted nitrate
    """
    V_MIN = 0.095
    V_MAX = 4.095
    a1 = (dac_min - dac_max) / (V_MIN - V_MAX)
    a0 = dac_max - V_MAX * a1

    nitrate = a1 * volts + a0

    if units == "mgNL":
        nitrate *= UMNO3_TO_MGNL

    return nitrate


def convert_ph_voltage_counts(ph_counts: np.ndarray):
    """Convert pH voltage counts to a floating point value

    :param ph_counts: pH voltage counts
    :return: pH voltage
    """
    ADC_VREF = 2.5
    GAIN = 1
    ADC_23BIT = 8388608
    ph_volts = ADC_VREF / GAIN * (ph_counts / ADC_23BIT - 1)
    return ph_volts


def _calculate_nernst(temperature: np.ndarray):
    """Calculate the nernst term using natual log

    :param temperature: temperature in kelvin
    :return: the nernst term (J/Coulomb; electrical potential; volts)
    """
    nernst_term = R * temperature / F * np.log(10)
    return nernst_term


def convert_internal_seafet_ph(
    ph_counts: np.ndarray,
    temperature: np.ndarray,
    coefs: PHSeaFETInternalCoefficients,
):
    """Calculates the internal pH on the total scale given the
    temperature and internal FET voltage

    :param ph_counts: pH voltage counts
    :param temperature: sample temperature
    :param coefs: SeaFET calibration coefficients
    :return: calculated pH on the total scale for the SeaFET internal
    reference
    """
    ph_volts = convert_ph_voltage_counts(ph_counts)

    # Eo(T) or temperature offset
    temperature_offset = coefs.int_k2 * temperature

    # Eo the cell reference voltage at in-situ conditions
    cell_ref_volts = coefs.int_k0 + temperature_offset
    nernst_term = _calculate_nernst(temperature + KELVIN_OFFSET_0C)
    ph = (ph_volts - cell_ref_volts) / nernst_term
    return ph


def convert_external_seafet_ph(
    ph_counts: np.ndarray,
    temperature: np.ndarray,
    salinity: np.ndarray,
    pressure: np.ndarray,
    coefs: PHSeaFETExternalCoefficients,
):
    """Calculates the external pH on the total scale given temperature,
    salinity, pressure and FET voltage counts

    :param ph_counts: ISFET external voltage counts
    :param temperature: sample temperature in Celsius
    :param salinity: sample salinity in psu
    :param pressure: sample pressure in dbar
    """

    TS_CR = 0.14  # relative concentration of sulfate in SW
    TS_MM = 96.062  # [g/mol] molar mass of sulfate
    CL_CR = 0.99889  # relative concentration of chloride in SW
    CL_MM = 35.453  # [g/mol] molar mass of chloride
    CL_TO_S = 1.80655  # [ppt, 10^{-3}] Chlorinity to Salinity

    def _molar_concentration(concentration: float, molar_mass: float, salinity: np.ndarray):
        """
        An estimate of constituent concentration in seawater is made from
        salinity on the basis of contancy of composition.

        :param concentration: Relative concentration of constituent
        :param molar_mass: Molar mass of constituent
        :param salinity: Salinity in PSU
        :return: molar concentration
        """
        return (concentration / molar_mass) * (salinity / CL_TO_S)

    def _molar_conc_chloride(salinity: np.ndarray):
        """Calculates the molar concentration of chloride

        :param salinity: Salinity in PSU
        :return: molar concentration of chloride
        """
        return _molar_concentration(CL_CR, CL_MM, salinity) * 1000 / (1000 - 1.005 * salinity)

    def _molar_conc_sulfate(salinity: np.ndarray):
        """Calculates the molar concentration of total sulfate

        :param salinity: Salinity in PSU
        :return: molar concentration of sulfate
        """
        return _molar_concentration(TS_CR, TS_MM, salinity)

    def _calculate_ionic_strength(salinity: np.ndarray):
        """Compute Salinity Ionic strength
        Ionic Strength (mol/kg H2O) from Dickson "Guide to Best Practices
        for Ocean CO2 Measurements", 2007, Chapter 5, page 11

        :param salinity: Salinity in PSU
        :return: Salinity ionic strength
        """
        C0 = 19.924
        C1 = 1000
        C2 = 1.005

        # 19.924*(S) / (1000 - 1.005*(S))
        ionic_strength = C0 * salinity / (C1 - C2 * salinity)
        return ionic_strength

    def _calculate_ks(
        temperature: np.ndarray,
        ionic_strength: np.ndarray,
        salinity: np.ndarray,
        pressure: np.ndarray,
    ):
        """Dissociation constant of sulfuric acid in seawater
        Dickson, A. G., J. Chemical Thermodynamics, 22:113-127, 1990
        The goodness of fit is .021.
        It was given in mol/kg-H2O. I convert it to mol/kg-SW.
        TYPO on p. 121: the constant e9 should be e8.
        This is from eqs 22 and 23 on p. 123, and Table 4 on p 121:

        :param temperature: Temperature in kelvin
        :param ionic_strength: Ionic strength
        :param salinity: Salinity in PSU
        :param pressure: Pressure in dbar
        :return: Dissociation constant of sulfuric acid in seawater
        """

        # *********** this should be re-tested
        C0 = -4276.1
        C1 = 141.328
        C2 = -23.093
        C3 = -13856
        C4 = 324.57
        C5 = -47.986
        C6 = 35474
        C7 = -771.54
        C8 = 114.723
        C9 = -2698
        C10 = 1776

        ln_ks = (
            C0 / temperature
            + C1
            + C2 * np.log(temperature)
            + (C3 / temperature + C4 + C5 * np.log(temperature)) * np.sqrt(ionic_strength)
            + (C6 / temperature + C7 + C8 * np.log(temperature)) * ionic_strength
            + (C9 / temperature) * np.sqrt(ionic_strength) * ionic_strength
            + (C10 / temperature) * np.pow(ionic_strength, 2)
        )

        # this is on the free pH scale in mol/kg-H2O
        khso4 = np.exp(ln_ks) * (1 - 0.001005 * salinity)  # convert to mol/kg-SW

        # UCI has two calculateKS functions. The first returns khso4 here,
        # but in practice the second is always used which adds the following

        temperature_C = temperature - KELVIN_OFFSET_0C
        # Partial molal volume and compressibility change for HSO4
        delta_vhso4 = -18.03 + 0.0466 * temperature_C + 0.000316 * temperature_C**2
        kappa_hso4 = (-4.53 + 0.09 * temperature_C) / 1000

        #  per Yui Press changed from dbar to bar here by / 10
        ln_khso4_fac = (
            (-delta_vhso4 + 0.5 * kappa_hso4 * (pressure / 10))
            * (pressure / 10)
            / (R * 10 * temperature)
        )

        #  bisulfate association constant at T, S, P
        khso4_tps = khso4 * np.exp(ln_khso4_fac)

        return khso4_tps

    def _calculate_adh(temperature):
        """Calculates the Debeye-Huckel constant (temperature [Celcius] dependence only)

        :param temperature: temperature in C
        :return: Debeye-Huckel constant
        """
        # This fit was made by Ken Johnson from data presented by:
        # Khoo et al. (Anal. Chem., 49, 29-34, 1977).
        C0 = 0.49172143
        C1 = 0.00067524
        # Modified from 0.00000343 to 0.0000034286,
        # email from Ken with newest version of MBARI code by Charles Branham 3/9/16
        C2 = 0.0000034286

        # 0.00000343*(t)*(t) + 0.00067524*(t) + 0.49172143
        adh = C0 + C1 * temperature + C2 * temperature**2

        return adh

    def _calculate_log_gamma_hcl(
        adh: np.ndarray, ionic_strength: np.ndarray, temperature: np.ndarray
    ):
        """
        Khoo et al. (Anal. Chem., 49, 29-34, 1977).
        \log \gamma_{\pm} \left( HCl\right) = \dfrac{-A \cdot \sqrt{I}}{1+\rho\cdot\sqrt{I}} +
        \left( B_{0} + B_{1}\cdot T \right) I
        As implemented in the calibration .xls

        :param adh: Debeye-Huckel constant
        :param ionic_strength: ionic strength
        :param temperature: Temperatur in C
        :return: Log Gamma HCL
        """
        # NOTE: There is a disagreement between Ken Johnson and Yui Takeshita
        # as to whether in-situ pressure correction is needed or already taken care of
        # by the ThermPress term being added to E0. In email correspondence
        # with Dave Murphy, Ken Johnson finally stated that no pressure correction
        # should be applied to Gamma_HCl

        rho = 1.394
        B0 = 0.08885
        B1 = 0.000111

        # As per instructions of KJ, should consider the data in
        # Dickson (J. Chem. Thermodynamics, 22, 113-127, 1990)
        log_gamma_hcl = (
            -1 * adh * np.sqrt(ionic_strength) / (1 + rho * np.sqrt(ionic_strength))
            + (B0 - B1 * temperature) * ionic_strength
        )

        return log_gamma_hcl

    def _calculate_thermal_pressure(temperature: np.ndarray, pressure: np.ndarray):
        """ThermPress at in-situ
        Thermpress = (-V_Cl x P + 0.5 K_Cl x P^2)/10F
        Where,
        P: Pressure in 'bar'
        V_Cl: Chloride partial molal volume (in cm^3 mol^-1)
        K_Cl: Chloride partial molal compressibility (in cm^3 mol^-1 bar^-1) (can be neglected [4, pg. 876])

        :param temperature: temperature in C
        :param pressure: pressure in dbar
        :return: thermal pressure
        """
        # // Partial molal volume and compressibility change for HCl from Millero
        delta_vhcl = 17.85 + 0.1044 * temperature - 0.001316 * temperature**2

        # // Pressure is in 'dbar' and need to divide by 10 to convert to 'bar'
        pressure_bar = pressure / 10

        # // Thermpress term
        thermal_pressure = -delta_vhcl * 0.0242 / (23061 * 1.01) * pressure_bar

        return thermal_pressure

    # Intermediate terms
    # tempK;   # Sample Temperature (K)
    temperature_k = temperature + KELVIN_OFFSET_0C
    # ST;        # Nernst Term
    st = _calculate_nernst(temperature=temperature_k)
    # mCl;       # Molar concentration of chloride
    mcl = _molar_conc_chloride(salinity=salinity)
    # TSO4;      # Molar concentration of total sulfate
    tso4 = _molar_conc_sulfate(salinity=salinity)
    # IonS;      # Salinity Ionic strength
    ionic_strength = _calculate_ionic_strength(salinity=salinity)

    adh = _calculate_adh(temperature=temperature)

    log_gamma_hcl = _calculate_log_gamma_hcl(
        adh=adh, ionic_strength=ionic_strength, temperature=temperature
    )

    thermal_pressure = _calculate_thermal_pressure(temperature=temperature, pressure=pressure)

    # // Dissociation constant of sulfuric acid in seawater
    # KSO4;      # Dissociation constant of sulfuric acid in seawater
    # this.KSO4 = calculateKS(this.tempK, this.IonS, salinity, tempC, pressure);
    kso4 = _calculate_ks(
        temperature=temperature_k,
        ionic_strength=ionic_strength,
        salinity=salinity,
        pressure=pressure,
    )

    ph_volts = convert_ph_voltage_counts(ph_counts)

    # Eo(T) or temperature offset
    eot = coefs.ext_k2 * temperature

    # Eo(P) or pressure offset
    eop = (
        coefs.f1 * pressure
        + coefs.f2 * pressure**2
        + coefs.f3 * pressure**3
        + coefs.f4 * pressure**4
        + coefs.f5 * pressure**5
        + coefs.f6 * pressure**6
    )

    # Calculate the External pH in Free Scale, firmware applies 2 factor here.
    # Convert pHfree(mol/kg H2O) to pHfree(mol/kg sln) per KSJ by using the
    # ratio (1000 - 1.005 * Salinity)/1000
    free_scale_ph = (
        (ph_volts - eot - eop - thermal_pressure - coefs.ext_k0) / st
        + np.log10(mcl)
        + (2 * log_gamma_hcl)
        - np.log10((1000 - 1.005 * salinity) / 1000)
    )

    total_ph = free_scale_ph - np.log10(1 + tso4 / kso4)

    return total_ph



"""Data processing unit tests.

"""

# Native imports
from logging import getLogger

# Third-party imports
import numpy as np
import pandas as pd
import pytest

# Sea-Bird imports

# Internal imports
import seabirdscientific.instrument_data as id
import seabirdscientific.processing as dp
import seabirdscientific.conversion as dc
from seabirdscientific.utils import close_enough
import seabirdscientific.instrument_data as id
import seabirdscientific.processing as dp
import seabirdscientific.conversion as dc
from seabirdscientific.utils import close_enough

logger = getLogger(__name__)


class TestLowPassFilter:
    expected_path = "./tests/resources/test-data/SBE37SM-filtered.asc"
    expected = pd.read_csv(expected_path)["Tv290C"].values

    source_path = "./tests/resources/test-data/SBE37SM-unfiltered.asc"
    source = pd.read_csv(source_path)["Tv290C"].values

    def test_low_pass_filter(self):
        filtered = dp.low_pass_filter(self.source, 10000, sample_interval=120)

        assert np.allclose(self.expected, filtered, rtol=0, atol=1e-4)


class TestAlignCtd:
    expected_data_path = "./tests/resources/test-data/SBE37SM-align.cnv"
    expected_data = id.cnv_to_instrument_data(expected_data_path)
    # expected data was algined with:
    #   sample_interval=120,
    #   tv290C + 12s
    #   cond0S/m + 150s
    #   prdM + 120s
    #   prdE + 0s
    #   sal00 - 240s

    source_data_path = "./tests/resources/test-data/SBE37SM.cnv"
    source_data = id.cnv_to_instrument_data(source_data_path)

    def test_align_ctd_add_simple_pass(self):
        expected_data_path = "./tests/resources/test-data/SBE37SM-align.cnv"
        expected_data = id.cnv_to_instrument_data(expected_data_path)

        source_data_path = "./tests/resources/test-data/SBE37SM.cnv"
        source_data = id.cnv_to_instrument_data(source_data_path).measurements["tv290C"].values

        # Fix last valid sample
        expected_data.measurements["tv290C"].values[
            len(expected_data.measurements["tv290C"].values) - 1
        ] = -9.99e-29
        result = dp.align_ctd(source_data, 12, 120)
        assert np.allclose(expected_data.measurements["tv290C"].values, result, atol=0.0001)

    def test_align_ctd_add_simple_pass_2(self):
        expected_data_path = "./tests/resources/test-data/SBE37SM-align.cnv"
        expected_data = id.cnv_to_instrument_data(expected_data_path)

        source_data_path = "./tests/resources/test-data/SBE37SM.cnv"
        source_data = id.cnv_to_instrument_data(source_data_path)

        # Fix last valid sample
        expected_data.measurements["cond0S/m"].values[
            len(expected_data.measurements["cond0S/m"].values) - 2
        ] = -9.99e-29
        result = dp.align_ctd(source_data.measurements["cond0S/m"].values, 150, 120)
        assert np.allclose(expected_data.measurements["cond0S/m"].values, result, atol=0.0001)

    def test_align_ctd_add_exact_factor(self):
        expected_data_path = "./tests/resources/test-data/SBE37SM-align.cnv"
        expected_data = id.cnv_to_instrument_data(expected_data_path)

        source_data_path = "./tests/resources/test-data/SBE37SM.cnv"
        source_data = id.cnv_to_instrument_data(source_data_path)

        result = dp.align_ctd(source_data.measurements["prdM"].values, 120, 120)
        assert np.allclose(expected_data.measurements["prdM"].values, result, atol=0.0001)

    def test_align_ctd_no_change(self):
        expected_data_path = "./tests/resources/test-data/SBE37SM-align.cnv"
        expected_data = id.cnv_to_instrument_data(expected_data_path)

        source_data_path = "./tests/resources/test-data/SBE37SM.cnv"
        source_data = id.cnv_to_instrument_data(source_data_path)

        result = dp.align_ctd(source_data.measurements["prdE"].values, 0, 120)
        assert np.allclose(expected_data.measurements["prdE"].values, result, atol=0.0001)

    def test_align_ctd_subtract(self):
        expected_data_path = "./tests/resources/test-data/SBE37SM-align.cnv"
        expected_data = id.cnv_to_instrument_data(expected_data_path)

        source_data_path = "./tests/resources/test-data/SBE37SM.cnv"
        source_data = id.cnv_to_instrument_data(source_data_path)

        # Fix last valid sample
        # expected_data.measurements['sal00'].values[len(expected_data.measurements['sal00'].values) - 2] = -9.99e-29

        result = dp.align_ctd(source_data.measurements["sal00"].values, -240, 120)
        assert np.allclose(expected_data.measurements["sal00"].values, result, atol=0.0001)


class TestCellThermalMass:
    def test_cell_thermal_mass_pass(self):
        expected_data_path = "./tests/resources/test-data/SBE37SM-ctm.cnv"
        expected_data = id.cnv_to_instrument_data(expected_data_path)
        source_data_path = "./tests/resources/test-data/SBE37SM.cnv"
        source_data = id.cnv_to_instrument_data(source_data_path)
        corrected_conductivity = dp.cell_thermal_mass(
            source_data.measurements["tv290C"].values,
            source_data.measurements["cond0S/m"].values,
            0.03,
            7.0,
            120.0,
        )
        assert np.allclose(
            expected_data.measurements["cond0S/m"].values,
            corrected_conductivity,
            atol=0.000001,
        )


class TestLoopEdit:
    def test_loop_edit_pressure_min_velocity_pass(self):
        expected_data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/CAST0002_mod_filt_loop_min_v.cnv"
        )
        data = id.cnv_to_instrument_data("./tests/resources/test-data/CAST0002_mod_filt.cnv")

        dp.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=dp.MinVelocityType.FIXED,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=False,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        assert mismatches / len(result_flags) < 0.01

    def test_loop_edit_pressure_min_velocity_remove_soak_pass(self):
        expected_data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/CAST0002_mod_filt_loop_min_v_remove_soak.cnv"
        )
        data = id.cnv_to_instrument_data("./tests/resources/test-data/CAST0002_mod_filt.cnv")

        dp.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=dp.MinVelocityType.FIXED,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=True,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        assert mismatches / len(result_flags) < 0.01

    def test_loop_edit_pressure_min_velocity_reset_flags_pass(self):
        expected_data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/CAST0002_mod_filt_loop_min_v.cnv"
        )
        data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/CAST0002_mod_filt_loop_min_v_remove_soak.cnv"
        )

        dp.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=dp.MinVelocityType.FIXED,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=False,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        assert mismatches / len(result_flags) < 0.01

    def test_loop_edit_pressure_min_velocity_exclude_flags_pass(self):
        expected_data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/CAST0002_mod_filt_loop_min_v_remove_soak.cnv"
        )
        data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/CAST0002_mod_filt_loop_min_v_exclude_flags_from_remove_soak.cnv"
        )

        dp.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=dp.MinVelocityType.FIXED,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=False,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=True,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        assert mismatches / len(result_flags) < 0.01

    def test_loop_edit_pressure_mean_speed_percent_remove_soak_pass(self):
        expected_data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/CAST0002_mod_filt_loop_percent_remove_soak.cnv"
        )
        data = id.cnv_to_instrument_data("./tests/resources/test-data/CAST0002_mod_filt.cnv")

        dp.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=dp.MinVelocityType.PERCENT,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=True,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        expected_flags = expected_data.measurements["flag"].values
        result_flags = data.measurements["flag"].values
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        # Less than 2% mismatched flags, outliers appear to conform to spec.
        assert mismatches / len(result_flags) < 0.02

    def test_loop_edit_pressure_mean_speed_percent_pass(self):
        expected_data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/CAST0002_mod_filt_loop_percent.cnv"
        )
        data = id.cnv_to_instrument_data("./tests/resources/test-data/CAST0002_mod_filt.cnv")

        dp.loop_edit_pressure(
            pressure=data.measurements["prSM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=dp.MinVelocityType.PERCENT,
            min_velocity=0.1,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=False,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=False,
            exclude_flags=False,
        )

        # SeaSoft is including earliest samples as local maxima during the first
        # downcast which would be discarded when requiring a minimum soak depth
        expected_flags = expected_data.measurements["flag"].values[100:]
        result_flags = data.measurements["flag"].values[100:]
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        # Less than 2% mismatched flags, outliers appear to conform to spec.
        assert mismatches / len(result_flags) < 0.02

    def test_loop_edit_pressure_min_velocity_pass_2(self):
        expected_data = id.cnv_to_instrument_data(
            "./tests/resources/test-data/SBE19plus_loop_edit.cnv"
        )
        data = id.cnv_to_instrument_data("./tests/resources/test-data/SBE19plus.cnv")

        dp.loop_edit_pressure(
            pressure=data.measurements["prdM"].values,
            latitude=data.latitude,
            flag=data.measurements["flag"].values,
            sample_interval=data.interval_s,
            min_velocity_type=dp.MinVelocityType.FIXED,
            min_velocity=0.25,
            window_size=3,
            mean_speed_percent=20,
            remove_surface_soak=True,
            min_soak_depth=5,
            max_soak_depth=20,
            use_deck_pressure_offset=True,
            exclude_flags=True,
        )

        # SeaSoft is including earliest samples as local maxima during the first
        # downcast which would be discarded when requiring a minimum soak depth
        expected_flags = expected_data.measurements["flag"].values[100:]
        result_flags = data.measurements["flag"].values[100:]
        mismatches = sum(expected_flags != result_flags)
        comp = pd.DataFrame({"expected": expected_flags, "result": result_flags})
        error_rows = []

        for n, row in comp.iterrows():
            if row["expected"] != row["result"]:
                error_rows.append(n)

        # Less than 2% mismatched flags, outliers appear to conform to spec.
        assert mismatches / len(result_flags) < 0.02


# class TestBinAverage:


class TestWildEdit:
    def test_wild_edit_pass(self):
        expected_dataset = id.cnv_to_instrument_data(
            "tests/resources/test-data/19plus_V2_CTD-processing_example_wild_edit.cnv"
        )
        expected_conductivity = expected_dataset.measurements["c0S/m"].values

        dataset = id.cnv_to_instrument_data(
            "tests/resources/test-data/19plus_V2_CTD-processing_example.cnv"
        )
        conductivity = dataset.measurements["c0S/m"].values
        flags = dataset.measurements["flag"].values

        wild_edit_output = dp.wild_edit(conductivity, flags, 2, 20, 100, 0, False)

        assert np.all(wild_edit_output == expected_conductivity)


# class TestWindowFilter:
#     # file_prefix = "C:\seabirdscientific\documentation\example_data\19plus_V2_CTD-processing_example_window_filter"
#     cnvdata = id.cnv_to_instrument_data(f"{file_prefix}.cnv")
#     pressure = cnvdata.measurements["prdM"].values
#     flags = cnvdata.measurements["flag"].values
#     window_width = 5
#     half_width = 1  # only applies to gaussian
#     offset = 0.25  # only applies to gaussian
#     sample_interval = cnvdata.interval_s  # only applies to gaussian

#     def test_boxcar_filter(self):
#         expected_dataset = id.cnv_to_instrument_data(f"{self.file_prefix}_boxcar_5.cnv")
#         expected_pressure = expected_dataset.measurements["prdM"].values

#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.BOXCAR,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             self.offset,
#             False,
#         )

#         assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

#     def test_boxcar_filter_exclude_flags(self):
#         expected_dataset = id.cnv_to_instrument_data(f"{self.file_prefix}_boxcar_5_excluded.cnv")
#         expected_pressure = expected_dataset.measurements["prdM"].values

#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.BOXCAR,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             self.offset,
#             True,
#         )

#         assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

#     def test_cosine_filter(self):
#         expected_dataset = id.cnv_to_instrument_data(f"{self.file_prefix}_cosine_5.cnv")
#         expected_pressure = expected_dataset.measurements["prdM"].values

#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.COSINE,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             self.offset,
#             False,
#         )

#         assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

#     def test_cosine_filter_exclude_flags(self):
#         expected_dataset = id.cnv_to_instrument_data(f"{self.file_prefix}_cosine_5_excluded.cnv")
#         expected_pressure = expected_dataset.measurements["prdM"].values

#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.COSINE,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             self.offset,
#             True,
#         )

#         assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

#     def test_triangle_filter(self):
#         expected_dataset = id.cnv_to_instrument_data(f"{self.file_prefix}_triangle_5.cnv")
#         expected_pressure = expected_dataset.measurements["prdM"].values

#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.TRIANGLE,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             self.offset,
#             False,
#         )

#         assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

#     def test_triangle_filter_exclude_flags(self):
#         expected_dataset = id.cnv_to_instrument_data(f"{self.file_prefix}_triangle_5_excluded.cnv")
#         expected_pressure = expected_dataset.measurements["prdM"].values

#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.TRIANGLE,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             self.offset,
#             True,
#         )

#         assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

#     def test_gaussian_filter(self):
#         expected_dataset = id.cnv_to_instrument_data(f"{self.file_prefix}_gaussian_5_1_025.cnv")
#         expected_pressure = expected_dataset.measurements["prdM"].values

#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.GAUSSIAN,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             self.offset,
#             False,
#         )

#         assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

#     def test_gaussian_filter_exclude_flags(self):
#         expected_dataset = id.cnv_to_instrument_data(f"{self.file_prefix}_gaussian_5_1_0_excluded.cnv")
#         expected_pressure = expected_dataset.measurements["prdM"].values

#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.GAUSSIAN,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             0,
#             True,
#         )

#         assert close_enough(filtered_pressure, expected_pressure, 3, 1e-12)

#     def test_median_filter_exclude_flags(self):
#         # median filter does not match SeaSoft
#         expected_pressure = [
#             7.405,
#             7.722,
#             7.98,
#             8.197,
#             8.286,
#             8.452,
#             8.654,
#             8.779,
#             8.879,
#             8.987,
#             9.169,
#             9.328,
#             9.415,
#             9.605,
#             9.818,
#             10.047,
#             10.305,
#             10.598,
#             10.917,
#             11.082,
#             11.418,
#             11.926,
#             12.265,
#             12.615,
#             12.936,
#             13.401,
#             13.681,
#             13.816,
#             14.051,
#             14.275,
#             14.481,
#             14.663,
#             14.839,
#             15.017,
#             15.192,
#             15.37,
#             15.556,
#             15.751,
#             15.852,
#             16.062,
#             16.408,
#             16.644,
#             16.903,
#             17.18,
#             17.608,
#             17.912,
#             18.068,
#             18.395,
#             18.735,
#             19.08,
#             19.418,
#         ]
#         filtered_pressure = dp.window_filter(
#             self.pressure,
#             self.flags,
#             dp.WindowFilterType.MEDIAN,
#             self.window_width,
#             self.cnvdata.interval_s,
#             self.half_width,
#             0,
#             True,
#         )
#         assert(close_enough(filtered_pressure, expected_pressure, 3, 1e-12))


class TestBuoyancy:
    # Testing data comes from a CalCOFI cruise
    temperature = np.asarray(
        [
            16.7373,
            16.5030,
            16.1106,
            14.3432,
            13.0211,
            12.0935,
            11.3933,
            11.2466,
            10.9219,
            10.4762,
            9.9460,
        ]
    )
    pressure = np.asarray([10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110])
    salinity = np.asarray(
        [
            33.2410,
            33.2321,
            33.2091,
            33.1329,
            33.0762,
            33.1391,
            33.2560,
            33.4015,
            33.5683,
            33.6766,
            33.7794,
        ]
    )
    expected_N2_win30 = np.asarray(
        [
            -9.990e-29,
            5.7702e-05,
            1.9197e-04,
            2.6735e-04,
            2.1888e-04,
            2.1620e-04,
            1.7374e-04,
            1.5761e-04,
            1.6905e-04,
            1.6099e-04,
            -9.990e-29,
        ]
    )
    expected_N_win30 = np.asarray(
        [-9.990e-29, 4.35, 7.94, 9.37, 8.48, 8.42, 7.55, 7.19, 7.45, 7.27, -9.990e-29]
    )
    expected_E_win30 = np.asarray(
        [
            -9.990e-29,
            5.8901e-06,
            1.9596e-05,
            2.7290e-05,
            2.2342e-05,
            2.2069e-05,
            1.7735e-05,
            1.6089e-05,
            1.7256e-05,
            1.6433e-05,
            -9.990e-29,
        ]
    )
    expected_E_pow_8_win30 = np.asarray(
        [
            -9.990e-29,
            589.0,
            1959.6,
            2729.0,
            2234.2,
            2206.9,
            1773.5,
            1608.9,
            1725.6,
            1643.3,
            -9.990e-29,
        ]
    )

    def test_modern_calc(self):
        # TODO: This test is failing. Fix as part of NSI-3061
        output_dataframe = dp.buoyancy(
            self.temperature,
            self.salinity,
            self.pressure,
            np.asarray([34.034167]),  # converted from metadata 34.02.03 N in H,M,S
            np.asarray([121.060556]),  # converted from metadata 121 03.38 W in H, M, S
            30,  # window size
            True,
        )
        print("output for modern calc:")
        print(output_dataframe)

        expected_dataframe = pd.DataFrame()
        expected_dataframe["N2"] = self.expected_N2_win30
        expected_dataframe["N"] = self.expected_N_win30
        expected_dataframe["E"] = self.expected_E_win30
        expected_dataframe["E10^-8"] = self.expected_E_pow_8_win30
        expected_dataframe["N2_diff"] = output_dataframe.N2 - self.expected_N2_win30
        expected_dataframe["N2_rpd"] = (
            100 * (output_dataframe.N2 - self.expected_N2_win30) / self.expected_N2_win30
        )

        print("Expected Dataframe")
        print(expected_dataframe)
        print(expected_dataframe.describe())

        # Comparing EOS-80 to TEOS-10 buoyancy calculations.
        # We do not expect them to agree better than +/-1.5% due to differences in the algorithms
        rel_tol = 0.015  # 1.5%
        assert output_dataframe["N2"].to_numpy() == pytest.approx(
            self.expected_N2_win30, rel=rel_tol
        )
        assert output_dataframe["N"].to_numpy() == pytest.approx(
            self.expected_N_win30, rel=rel_tol
        )
        assert output_dataframe["E"].to_numpy() == pytest.approx(
            self.expected_E_win30, rel=rel_tol
        )
        assert output_dataframe["E10^-8"].to_numpy() == pytest.approx(
            self.expected_E_pow_8_win30, rel=rel_tol
        )

    def test_retro_calc(self):
        # TODO: This test is failing. Fix as part of NSI-3061
        output_dataframe = dp.buoyancy(
            self.temperature,
            self.salinity,
            self.pressure,
            np.asarray([34.034167]),  # converted from metadata 34.02.03 N in H,M,S
            np.asarray([121.060556]),  # converted from metadata 121 03.38 W in H, M, S
            30,  # window size
            False,
        )

        print("output for retro calc:")
        print(output_dataframe)

        expected_dataframe = pd.DataFrame()
        expected_dataframe["N2"] = self.expected_N2_win30
        expected_dataframe["N"] = self.expected_N_win30
        expected_dataframe["E"] = self.expected_E_win30
        expected_dataframe["E10^-8"] = self.expected_E_pow_8_win30
        expected_dataframe["N2_diff"] = output_dataframe.N2 - self.expected_N2_win30
        expected_dataframe["N2_rpd"] = (
            100 * (output_dataframe.N2 - self.expected_N2_win30) / self.expected_N2_win30
        )

        print("Expected Dataframe")
        print(expected_dataframe)
        print(expected_dataframe.describe())

        # Comparing SBE Data Processing C++ to local Python results using the same EOS-80 calculations.
        # We expect very very close agreement: << 1% differnce
        rel_tol = 0.0026  # 0.26%
        assert output_dataframe["N2"].to_numpy() == pytest.approx(
            self.expected_N2_win30, rel=rel_tol
        )
        assert output_dataframe["N"].to_numpy() == pytest.approx(
            self.expected_N_win30, rel=rel_tol
        )
        assert output_dataframe["E"].to_numpy() == pytest.approx(
            self.expected_E_win30, rel=rel_tol
        )
        assert output_dataframe["E10^-8"].to_numpy() == pytest.approx(
            self.expected_E_pow_8_win30, rel=rel_tol
        )


class TestNitrate:
    dac_min = -5
    dac_max = 100
    voltages = np.array([0.095, 1, 2, 3, 4.095])

    def test_convert_nitrate_umno3(self):
        expected_umno3 = np.array([-5, 18.75625, 45.00625, 71.25625, 100])
        nitrate = dc.convert_nitrate(self.voltages, dac_min=self.dac_min, dac_max=self.dac_max)
        assert np.allclose(expected_umno3, nitrate, atol=0.000001)

    def test_convert_nitrate_mgnl(self):
        expected_mgnl = np.array([-0.070035, 0.26271879375, 0.63040254375, 0.99808629375, 1.4007])
        nitrate = dc.convert_nitrate(self.voltages, dac_min=self.dac_min, dac_max=self.dac_max, units='mgNL')
        assert np.allclose(expected_mgnl, nitrate, atol=0.000001)